from typing import Any, Dict, cast, List, Iterable, Tuple, Optional
from typing_extensions import Final
from dataclasses import asdict

from opentrons.hardware_control.types import OT3AxisKind, InstrumentProbeType
from .types import (
    OT3Config,
    ByGantryLoad,
    OT3CurrentSettings,
    OT3MotionSettings,
    OT3Transform,
    Offset,
    OT3CalibrationSettings,
    CapacitivePassSettings,
    LiquidProbeSettings,
    ZSenseSettings,
    EdgeSenseSettings,
)


DEFAULT_PIPETTE_OFFSET = [0.0, 0.0, 0.0]
DEFAULT_MODULE_OFFSET = [0.0, 0.0, 0.0]

DEFAULT_LIQUID_PROBE_SETTINGS: Final[LiquidProbeSettings] = LiquidProbeSettings(
    mount_speed=5,
    plunger_speed=15,
    plunger_impulse_time=0.2,
    sensor_threshold_pascals=15,
    aspirate_while_sensing=False,
    z_overlap_between_passes_mm=0.1,
    plunger_reset_offset=2.0,
    samples_for_baselining=20,
    sample_time_sec=0.004,
)

DEFAULT_CALIBRATION_SETTINGS: Final[OT3CalibrationSettings] = OT3CalibrationSettings(
    z_offset=ZSenseSettings(
        pass_settings=CapacitivePassSettings(
            prep_distance_mm=4.0,
            max_overrun_distance_mm=5.0,
            speed_mm_per_s=1.0,
            sensor_threshold_pf=3.0,
        ),
    ),
    edge_sense=EdgeSenseSettings(
        overrun_tolerance_mm=0.4,
        early_sense_tolerance_mm=0.5,
        pass_settings=CapacitivePassSettings(
            prep_distance_mm=1,
            max_overrun_distance_mm=0.5,
            speed_mm_per_s=1,
            sensor_threshold_pf=3.0,
        ),
        search_initial_tolerance_mm=12.0,
        search_iteration_limit=8,
    ),
    probe_length=44.5,
)

ROBOT_CONFIG_VERSION: Final = 1
DEFAULT_LOG_LEVEL: Final = "INFO"
DEFAULT_MACHINE_TRANSFORM: Final[OT3Transform] = [
    [-1.0, 0.0, 0.0],
    [0.0, -1.0, 0.0],
    [0.0, 0.0, -1.0],
]
DEFAULT_BELT_ATTITUDE: Final[OT3Transform] = [
    [1.0, 0.0, 0.0],
    [0.0, 1.0, 0.0],
    [0.0, 0.0, 1.0],
]
DEFAULT_CARRIAGE_OFFSET: Final[Offset] = (477.20, 493.8, 253.475)
DEFAULT_LEFT_MOUNT_OFFSET: Final[Offset] = (-13.5, -60.5, 255.675)
DEFAULT_RIGHT_MOUNT_OFFSET: Final[Offset] = (40.5, -60.5, 255.675)
DEFAULT_GRIPPER_MOUNT_OFFSET: Final[Offset] = (84.55, -12.75, 93.85)
DEFAULT_SAFE_HOME_DISTANCE: Final = 5
DEFAULT_CALIBRATION_AXIS_MAX_SPEED: Final = 30

DEFAULT_MAX_SPEEDS: Final[ByGantryLoad[Dict[OT3AxisKind, float]]] = ByGantryLoad(
    high_throughput={
        OT3AxisKind.X: 350,
        OT3AxisKind.Y: 300,
        OT3AxisKind.Z: 35,
        OT3AxisKind.P: 15,
        OT3AxisKind.Z_G: 50,
        OT3AxisKind.Q: 10,
    },
    low_throughput={
        OT3AxisKind.X: 350,
        OT3AxisKind.Y: 300,
        OT3AxisKind.Z: 100,
        OT3AxisKind.P: 70,
        OT3AxisKind.Z_G: 50,
    },
)

DEFAULT_ACCELERATIONS: Final[ByGantryLoad[Dict[OT3AxisKind, float]]] = ByGantryLoad(
    high_throughput={
        OT3AxisKind.X: 700,
        OT3AxisKind.Y: 600,
        OT3AxisKind.Z: 150,
        OT3AxisKind.P: 30,
        OT3AxisKind.Z_G: 150,
        OT3AxisKind.Q: 10,
    },
    low_throughput={
        OT3AxisKind.X: 800,
        OT3AxisKind.Y: 600,
        OT3AxisKind.Z: 150,
        OT3AxisKind.P: 100,
        OT3AxisKind.Z_G: 150,
    },
)

DEFAULT_MAX_SPEED_DISCONTINUITY: Final[
    ByGantryLoad[Dict[OT3AxisKind, float]]
] = ByGantryLoad(
    high_throughput={
        OT3AxisKind.X: 10,
        OT3AxisKind.Y: 10,
        OT3AxisKind.Z: 5,
        OT3AxisKind.P: 5,
        OT3AxisKind.Z_G: 5,
        OT3AxisKind.Q: 5,
    },
    low_throughput={
        OT3AxisKind.X: 10,
        OT3AxisKind.Y: 10,
        OT3AxisKind.Z: 5,
        OT3AxisKind.P: 10,
        OT3AxisKind.Z_G: 5,
    },
)

DEFAULT_DIRECTION_CHANGE_SPEED_DISCONTINUITY: Final[
    ByGantryLoad[Dict[OT3AxisKind, float]]
] = ByGantryLoad(
    high_throughput={
        OT3AxisKind.X: 5,
        OT3AxisKind.Y: 5,
        OT3AxisKind.Z: 1,
        OT3AxisKind.P: 5,
        OT3AxisKind.Q: 5,
        OT3AxisKind.Z_G: 5,
    },
    low_throughput={
        OT3AxisKind.X: 5,
        OT3AxisKind.Y: 5,
        OT3AxisKind.Z: 1,
        OT3AxisKind.P: 5,
        OT3AxisKind.Z_G: 5,
    },
)

DEFAULT_HOLD_CURRENT: Final[ByGantryLoad[Dict[OT3AxisKind, float]]] = ByGantryLoad(
    high_throughput={
        OT3AxisKind.X: 0.5,
        OT3AxisKind.Y: 0.5,
        OT3AxisKind.Z: 0.5,
        OT3AxisKind.P: 0.3,
        OT3AxisKind.Z_G: 0.2,
        OT3AxisKind.Q: 0.3,
    },
    low_throughput={
        OT3AxisKind.X: 0.5,
        OT3AxisKind.Y: 0.5,
        OT3AxisKind.Z: 0.1,
        OT3AxisKind.P: 0.3,
        OT3AxisKind.Z_G: 0.2,
    },
)

DEFAULT_RUN_CURRENT: Final[ByGantryLoad[Dict[OT3AxisKind, float]]] = ByGantryLoad(
    high_throughput={
        OT3AxisKind.X: 1.25,
        OT3AxisKind.Y: 1.4,
        OT3AxisKind.Z: 1.5,
        OT3AxisKind.P: 0.8,
        OT3AxisKind.Z_G: 0.67,
        OT3AxisKind.Q: 1.5,
    },
    low_throughput={
        OT3AxisKind.X: 1.25,
        OT3AxisKind.Y: 1.2,
        OT3AxisKind.Z: 1.0,
        # TODO: verify this value
        OT3AxisKind.P: 1.0,
        OT3AxisKind.Z_G: 0.67,
    },
)


def _build_log_files_with_default(
    from_conf: Any,
    default: Optional[Dict[InstrumentProbeType, str]],
) -> Optional[Dict[InstrumentProbeType, str]]:
    print(f"from_conf {from_conf} default {default}")
    if not isinstance(from_conf, dict):
        if default is None:
            return None
        else:
            return {k: v for k, v in default.items()}
    else:
        validated: Dict[InstrumentProbeType, str] = {}
        for k, v in from_conf.items():
            if isinstance(k, InstrumentProbeType):
                validated[k] = v
            else:
                try:
                    enumval = InstrumentProbeType[k]
                except KeyError:  # not an enum entry
                    pass
                else:
                    validated[enumval] = v
        print(f"result {validated}")
        return validated


def _build_dict_with_default(
    from_conf: Any,
    default: Dict[OT3AxisKind, float],
) -> Dict[OT3AxisKind, float]:
    if not isinstance(from_conf, dict):
        return {k: v for k, v in default.items()}
    else:
        validated: Dict[OT3AxisKind, float] = {}
        # Keep what is specified, handling it being
        # either enum element name string or enum element directly
        for k, v in from_conf.items():
            if isinstance(k, OT3AxisKind):
                validated[k] = v
            else:
                try:
                    enumval = OT3AxisKind[k]
                except KeyError:  # not an enum entry
                    pass
                else:
                    validated[enumval] = v
        # Add what's missing relative to the default
        for k, default_v in default.items():
            if k in from_conf:
                validated[k] = from_conf[k]
            elif k.name in from_conf:
                validated[k] = from_conf[k.name]
            else:
                validated[k] = default_v
        return validated


def _build_default_bpk(
    from_conf: Any, default: ByGantryLoad[Dict[OT3AxisKind, float]]
) -> ByGantryLoad[Dict[OT3AxisKind, float]]:
    return ByGantryLoad(
        low_throughput=_build_dict_with_default(
            from_conf.get("low_throughput", {}), default.low_throughput
        ),
        high_throughput=_build_dict_with_default(
            from_conf.get("high_throughput", {}), default.high_throughput
        ),
    )


def _build_default_offset(from_conf: Any, default: Offset) -> Offset:
    if not isinstance(from_conf, (list, tuple)) or len(from_conf) != 3:
        return default
    return cast(Offset, tuple(from_conf))


def _build_default_transform(
    from_conf: Any, default: List[List[float]]
) -> List[List[float]]:
    if (
        not isinstance(from_conf, list)
        or len(from_conf) != 3
        or not all(isinstance(elem, list) for elem in from_conf)
        or not all(len(e) == 3 for e in from_conf)
        or not all(
            all(isinstance(elem, (int, float)) for elem in vec) for vec in from_conf
        )
    ):
        return default
    return cast(OT3Transform, from_conf)


def _build_default_cap_pass(
    from_conf: Any, default: CapacitivePassSettings
) -> CapacitivePassSettings:
    return CapacitivePassSettings(
        prep_distance_mm=from_conf.get("prep_distance_mm", default.prep_distance_mm),
        max_overrun_distance_mm=from_conf.get(
            "max_overrun_distance_mm", default.max_overrun_distance_mm
        ),
        speed_mm_per_s=from_conf.get("speed_mm_per_s", default.speed_mm_per_s),
        sensor_threshold_pf=from_conf.get(
            "sensor_threshold_pf", default.sensor_threshold_pf
        ),
    )


def _build_default_liquid_probe(
    from_conf: Any, default: LiquidProbeSettings
) -> LiquidProbeSettings:
    return LiquidProbeSettings(
        mount_speed=from_conf.get("mount_speed", default.mount_speed),
        plunger_speed=from_conf.get("plunger_speed", default.plunger_speed),
        plunger_impulse_time=from_conf.get(
            "plunger_impulse_time", default.plunger_impulse_time
        ),
        sensor_threshold_pascals=from_conf.get(
            "sensor_threshold_pascals", default.sensor_threshold_pascals
        ),
        aspirate_while_sensing=from_conf.get(
            "aspirate_while_sensing", default.aspirate_while_sensing
        ),
        z_overlap_between_passes_mm=from_conf.get(
            "z_overlap_between_passes_mm", default.z_overlap_between_passes_mm
        ),
        plunger_reset_offset=from_conf.get(
            "plunger_reset_offset", default.plunger_reset_offset
        ),
        samples_for_baselining=from_conf.get(
            "samples_for_baselining", default.samples_for_baselining
        ),
        sample_time_sec=from_conf.get("sample_time_sec", default.sample_time_sec),
    )


def _build_default_z_pass(from_conf: Any, default: ZSenseSettings) -> ZSenseSettings:
    return ZSenseSettings(
        pass_settings=_build_default_cap_pass(
            from_conf.get("pass_settings", {}), default.pass_settings
        ),
    )


def _build_default_edge_sense(
    from_conf: Any, default: EdgeSenseSettings
) -> EdgeSenseSettings:
    return EdgeSenseSettings(
        overrun_tolerance_mm=from_conf.get(
            "overrun_tolerance_mm", default.overrun_tolerance_mm
        ),
        early_sense_tolerance_mm=from_conf.get(
            "early_sense_tolerance_mm", default.early_sense_tolerance_mm
        ),
        pass_settings=_build_default_cap_pass(
            from_conf.get("pass_settings", {}), default.pass_settings
        ),
        search_initial_tolerance_mm=from_conf.get(
            "search_initial_tolerance_mm", default.search_initial_tolerance_mm
        ),
        search_iteration_limit=from_conf.get(
            "search_iteration_limit", default.search_iteration_limit
        ),
    )


def _build_default_calibration(
    from_conf: Any, default: OT3CalibrationSettings
) -> OT3CalibrationSettings:
    return OT3CalibrationSettings(
        z_offset=_build_default_z_pass(from_conf.get("z_offset", {}), default.z_offset),
        edge_sense=_build_default_edge_sense(
            from_conf.get("edge_sense", {}), default.edge_sense
        ),
        probe_length=from_conf.get("probe_length", default.probe_length),
    )


def build_with_defaults(robot_settings: Dict[str, Any]) -> OT3Config:
    motion_settings = robot_settings.get("motion_settings", {})
    current_settings = robot_settings.get("current_settings", {})
    return OT3Config(
        model="OT-3 Standard",
        version=ROBOT_CONFIG_VERSION,
        name=robot_settings.get("name", "Grace Hopper"),
        log_level=robot_settings.get("log_level", DEFAULT_LOG_LEVEL),
        motion_settings=OT3MotionSettings(
            default_max_speed=_build_default_bpk(
                motion_settings.get("default_max_speed", {}), DEFAULT_MAX_SPEEDS
            ),
            acceleration=_build_default_bpk(
                motion_settings.get("acceleration", {}), DEFAULT_ACCELERATIONS
            ),
            max_speed_discontinuity=_build_default_bpk(
                motion_settings.get("max_speed_discontinuity", {}),
                DEFAULT_MAX_SPEED_DISCONTINUITY,
            ),
            direction_change_speed_discontinuity=_build_default_bpk(
                motion_settings.get("direction_change_speed_discontinuity", {}),
                DEFAULT_DIRECTION_CHANGE_SPEED_DISCONTINUITY,
            ),
        ),
        current_settings=OT3CurrentSettings(
            hold_current=_build_default_bpk(
                current_settings.get("hold_current", {}),
                DEFAULT_HOLD_CURRENT,
            ),
            run_current=_build_default_bpk(
                current_settings.get("run_current", {}),
                DEFAULT_RUN_CURRENT,
            ),
        ),
        safe_home_distance=robot_settings.get(
            "safe_home_distance", DEFAULT_SAFE_HOME_DISTANCE
        ),
        deck_transform=_build_default_transform(
            robot_settings.get("deck_transform", []), DEFAULT_MACHINE_TRANSFORM
        ),
        carriage_offset=_build_default_offset(
            robot_settings.get("carriage_offset", []), DEFAULT_CARRIAGE_OFFSET
        ),
        left_mount_offset=_build_default_offset(
            robot_settings.get("left_mount_offset", []), DEFAULT_LEFT_MOUNT_OFFSET
        ),
        right_mount_offset=_build_default_offset(
            robot_settings.get("right_mount_offset", []), DEFAULT_RIGHT_MOUNT_OFFSET
        ),
        gripper_mount_offset=_build_default_offset(
            robot_settings.get("gripper_mount_offset", []), DEFAULT_GRIPPER_MOUNT_OFFSET
        ),
        calibration=_build_default_calibration(
            robot_settings.get("calibration", {}), DEFAULT_CALIBRATION_SETTINGS
        ),
        liquid_sense=_build_default_liquid_probe(
            robot_settings.get("liquid_sense", {}), DEFAULT_LIQUID_PROBE_SETTINGS
        ),
    )


def serialize(config: OT3Config) -> Dict[str, Any]:
    def _build_dict(pairs: Iterable[Tuple[Any, Any]]) -> Dict[str, Any]:
        def _normalize_key(key: Any) -> Any:
            if isinstance(key, OT3AxisKind) or isinstance(key, InstrumentProbeType):
                return key.name
            return key

        def _normalize_value(value: Any) -> Any:
            if isinstance(value, dict):
                return {
                    _normalize_key(k): _normalize_value(v) for k, v in value.items()
                }
            else:
                return value

        return dict((_normalize_key(key), _normalize_value(val)) for key, val in pairs)

    return asdict(config, dict_factory=_build_dict)



"""A set that preserves the order in which elements are added."""

from __future__ import annotations

from typing import (
    Dict,
    Generic,
    Hashable,
    Iterable,
    Iterator,
    Set,
    TypeVar,
    Union,
    overload,
)
from typing_extensions import Literal


_SetElementT = TypeVar("_SetElementT", bound=Hashable)

_DefaultValueT = TypeVar("_DefaultValueT")


class _NOT_SPECIFIED:
    """Value not specified sentinel."""


# Implemented as a standalone class for clarity.
# If this proves insufficient, we can get many methods for free
# by subclassing collections.abc.MutableSet.
class OrderedSet(Generic[_SetElementT]):
    """A set that preserves the order in which elements are added.

    Args:
        source_iterable: An ordered iterable of initial elements.
    """

    def __init__(self, source_iterable: Iterable[_SetElementT] = tuple()) -> None:
        self._elements: Dict[_SetElementT, Literal[True]] = {}

        for element in source_iterable:
            self.add(element)

    def add(self, element: _SetElementT) -> None:
        """Add ``element`` to the set.

        If ``element`` is already in the set, it is not added again,
        and its existing position in the set is retained.
        """
        self._elements[element] = True

    def remove(self, element: _SetElementT) -> None:
        """Remove ``element`` from the set.

        Raises:
            KeyError: If ``element`` is not in the set.
        """
        del self._elements[element]

    def discard(self, element: _SetElementT) -> None:
        """Remove ``element`` from the set, if it is present."""
        try:
            self.remove(element)
        except KeyError:
            pass

    def clear(self) -> None:
        """Remove all elements from the set."""
        self._elements.clear()

    @overload
    def head(self) -> _SetElementT:
        ...

    @overload
    def head(
        self, default_value: _DefaultValueT
    ) -> Union[_SetElementT, _DefaultValueT]:
        ...

    def head(
        self, default_value: Union[_DefaultValueT, _NOT_SPECIFIED] = _NOT_SPECIFIED()
    ) -> Union[_SetElementT, _DefaultValueT]:
        """Get the head (oldest-added element) of the set.

        Args:
            default_value: A value to return if set is empty.

        Returns:
            The head of the set, or the default value, if specified.

        Raises:
            IndexError: set is empty and default was not specified.
        """
        try:
            return next(iter(self._elements))
        except StopIteration:
            if isinstance(default_value, _NOT_SPECIFIED):
                raise IndexError("Set is empty") from None
            else:
                return default_value

    def __iter__(self) -> Iterator[_SetElementT]:
        """Enable iteration over all elements in the set.

        Elements are returned in the order they were added, oldest first.
        """
        return iter(self._elements)

    def __len__(self) -> int:
        """Return the number of unique elements added to the set."""
        return len(self._elements)

    def __eq__(self, other: object) -> bool:
        """Return whether this set is equal to another object.

        True if the other object is also an ordered set,
        and it contains the same elements in the same order.
        """
        if isinstance(other, OrderedSet):
            return list(self) == list(other)
        else:
            return False

    def __sub__(
        self, other: Union[OrderedSet[_SetElementT], Set[_SetElementT]]
    ) -> OrderedSet[_SetElementT]:
        """Return this set, without any elements that appear in `other`.

        The elements that aren't removed retain their original relative order.
        """
        return OrderedSet(e for e in self if e not in other)

    def __repr__(self) -> str:  # noqa: D105
        # Use repr() on the keys view in case it's super long and Python is smart
        # enough to abbreviate it.
        elements_str = repr(self._elements.keys())
        return f"OrderedSet({elements_str})"



"""Protocol engine state management."""
from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Dict, List, Optional, Sequence, TypeVar
from typing_extensions import ParamSpec

from opentrons_shared_data.deck.types import DeckDefinitionV5
from opentrons_shared_data.robot.types import RobotDefinition

from opentrons.protocol_engine.error_recovery_policy import ErrorRecoveryPolicy
from opentrons.protocol_engine.types import ModuleOffsetData
from opentrons.util.change_notifier import ChangeNotifier

from ..resources import DeckFixedLabware
from ..actions import Action, ActionHandler
from ._abstract_store import HasState, HandlesActions
from .commands import CommandState, CommandStore, CommandView
from .addressable_areas import (
    AddressableAreaState,
    AddressableAreaStore,
    AddressableAreaView,
)
from .labware import LabwareState, LabwareStore, LabwareView
from .pipettes import PipetteState, PipetteStore, PipetteView
from .modules import ModuleState, ModuleStore, ModuleView
from .liquids import LiquidState, LiquidView, LiquidStore
from .tips import TipState, TipView, TipStore
from .wells import WellState, WellView, WellStore
from .geometry import GeometryView
from .motion import MotionView
from .files import FileView, FileState, FileStore
from .config import Config
from .state_summary import StateSummary
from ..types import DeckConfigurationType


_ParamsT = ParamSpec("_ParamsT")
_ReturnT = TypeVar("_ReturnT")


@dataclass(frozen=True)
class State:
    """Underlying engine state."""

    commands: CommandState
    addressable_areas: AddressableAreaState
    labware: LabwareState
    pipettes: PipetteState
    modules: ModuleState
    liquids: LiquidState
    tips: TipState
    wells: WellState
    files: FileState


class StateView(HasState[State]):
    """A read-only view of computed state."""

    _state: State
    _commands: CommandView
    _addressable_areas: AddressableAreaView
    _labware: LabwareView
    _pipettes: PipetteView
    _modules: ModuleView
    _liquid: LiquidView
    _tips: TipView
    _wells: WellView
    _geometry: GeometryView
    _motion: MotionView
    _files: FileView
    _config: Config

    @property
    def commands(self) -> CommandView:
        """Get state view selectors for commands state."""
        return self._commands

    @property
    def addressable_areas(self) -> AddressableAreaView:
        """Get state view selectors for addressable area state."""
        return self._addressable_areas

    @property
    def labware(self) -> LabwareView:
        """Get state view selectors for labware state."""
        return self._labware

    @property
    def pipettes(self) -> PipetteView:
        """Get state view selectors for pipette state."""
        return self._pipettes

    @property
    def modules(self) -> ModuleView:
        """Get state view selectors for hardware module state."""
        return self._modules

    @property
    def liquid(self) -> LiquidView:
        """Get state view selectors for liquid state."""
        return self._liquid

    @property
    def tips(self) -> TipView:
        """Get state view selectors for tip state."""
        return self._tips

    @property
    def wells(self) -> WellView:
        """Get state view selectors for well state."""
        return self._wells

    @property
    def geometry(self) -> GeometryView:
        """Get state view selectors for derived geometry state."""
        return self._geometry

    @property
    def motion(self) -> MotionView:
        """Get state view selectors for derived motion state."""
        return self._motion

    @property
    def files(self) -> FileView:
        """Get state view selectors for engine create file state."""
        return self._files

    @property
    def config(self) -> Config:
        """Get ProtocolEngine configuration."""
        return self._config

    def get_summary(self) -> StateSummary:
        """Get protocol run data."""
        error = self._commands.get_error()
        # TODO maybe add summary here for AA
        return StateSummary.construct(
            status=self._commands.get_status(),
            errors=[] if error is None else [error],
            pipettes=self._pipettes.get_all(),
            labware=self._labware.get_all(),
            labwareOffsets=self._labware.get_labware_offsets(),
            modules=self._modules.get_all(),
            completedAt=self._state.commands.run_completed_at,
            startedAt=self._state.commands.run_started_at,
            liquids=self._liquid.get_all(),
            wells=self._wells.get_all(),
            hasEverEnteredErrorRecovery=self._commands.get_has_entered_recovery_mode(),
            files=self._state.files.file_ids,
        )


class StateStore(StateView, ActionHandler):
    """ProtocolEngine state store.

    A StateStore manages several substores, which will modify themselves in
    reaction to commands and other protocol events. State instances inside
    stores should be treated as immutable.
    """

    def __init__(
        self,
        *,
        config: Config,
        deck_definition: DeckDefinitionV5,
        deck_fixed_labware: Sequence[DeckFixedLabware],
        robot_definition: RobotDefinition,
        is_door_open: bool,
        error_recovery_policy: ErrorRecoveryPolicy,
        change_notifier: Optional[ChangeNotifier] = None,
        module_calibration_offsets: Optional[Dict[str, ModuleOffsetData]] = None,
        deck_configuration: Optional[DeckConfigurationType] = None,
        notify_publishers: Optional[Callable[[], None]] = None,
    ) -> None:
        """Initialize a StateStore and its substores.

        Arguments:
            config: Top-level configuration.
            deck_definition: The deck definition to preload into
                labware state.
            deck_fixed_labware: Labware definitions from the deck
                definition to preload into labware state.
            is_door_open: Whether the robot's door is currently open.
            error_recovery_policy: The run's initial error recovery policy.
            change_notifier: Internal state change notifier.
            module_calibration_offsets: Module offsets to preload.
            deck_configuration: The initial deck configuration the addressable area store will be instantiated with.
            robot_definition: Static information about the robot type being used.
            notify_publishers: Notifies robot server publishers of internal state change.
        """
        self._command_store = CommandStore(
            config=config,
            is_door_open=is_door_open,
            error_recovery_policy=error_recovery_policy,
        )
        self._pipette_store = PipetteStore()
        if deck_configuration is None:
            deck_configuration = []
        self._addressable_area_store = AddressableAreaStore(
            deck_configuration=deck_configuration,
            config=config,
            deck_definition=deck_definition,
            robot_definition=robot_definition,
        )
        self._labware_store = LabwareStore(
            deck_fixed_labware=deck_fixed_labware,
            deck_definition=deck_definition,
        )
        self._module_store = ModuleStore(
            config=config,
            deck_fixed_labware=deck_fixed_labware,
            module_calibration_offsets=module_calibration_offsets,
        )
        self._liquid_store = LiquidStore()
        self._tip_store = TipStore()
        self._well_store = WellStore()
        self._file_store = FileStore()

        self._substores: List[HandlesActions] = [
            self._command_store,
            self._pipette_store,
            self._addressable_area_store,
            self._labware_store,
            self._module_store,
            self._liquid_store,
            self._tip_store,
            self._well_store,
            self._file_store,
        ]
        self._config = config
        self._change_notifier = change_notifier or ChangeNotifier()
        self._notify_robot_server = notify_publishers
        self._initialize_state()

    def handle_action(self, action: Action) -> None:
        """Modify State in reaction to an action.

        Arguments:
            action: An action object representing a state change. Will be
                passed to all substores so they can react accordingly.
        """
        for substore in self._substores:
            substore.handle_action(action)

        self._update_state_views()

    async def wait_for(
        self,
        condition: Callable[_ParamsT, _ReturnT],
        *args: _ParamsT.args,
        **kwargs: _ParamsT.kwargs,
    ) -> _ReturnT:
        """Wait for a condition to become true, checking whenever state changes.

        If the condition is already true, return immediately.

        !!! Warning:
            This will only return when `condition` is true right now.
            If you're not careful, this can cause you to miss updates,
            and potentially wait forever.

            For example, suppose:

            1. Things start out in state A.
            2. You start waiting specifically for state B.
            3. State transitions A -> B.
               Your task could theoretically run now,
               but the event loop decides not to do that.
            4. State transitions B -> C.
            5. The event loop decides to run your task now.
               But this method will only return when the state is B,
               which it isn't now, so your task will stay waiting indefinitely.

            To fix this, design your `condition` to look for something that,
            after it first becomes true, remains true forever. For example,
            suppose you want to wait for the engine to reach a specific command.
            Don't use a `condition` that checks if that command is running now.
            Instead, use a `condition` that checks if it ever has been running.

            If this isn't possible, then you need other means to ensure that
            no other task can transition the state out of the condition that you
            care about until after you've had a chance to operate on the state
            while it's in that condition.

        Arguments:
            condition: A function that returns a truthy value when the `await`
                should resolve.
            *args: Positional arguments to pass to `condition`.
            **kwargs: Named arguments to pass to `condition`.

        Returns:
            The truthy value returned by the `condition` function.

        Raises:
            The exception raised by the `condition` function, if any.
        """

        def predicate() -> _ReturnT:
            return condition(*args, **kwargs)

        return await self._wait_for(condition=predicate, truthiness_to_wait_for=True)

    async def wait_for_not(
        self,
        condition: Callable[_ParamsT, _ReturnT],
        *args: _ParamsT.args,
        **kwargs: _ParamsT.kwargs,
    ) -> _ReturnT:
        """Like `wait_for()`, except wait for the condition to become false.

        See the documentation in `wait_for()`, especially the warning about condition
        design.

        The advantage of having this separate method over just passing a wrapper lambda
        as the condition to `wait_for()` yourself is that wrapper lambdas are hard to
        test in the mock-heavy Decoy + Protocol Engine style.
        """

        def predicate() -> _ReturnT:
            return condition(*args, **kwargs)

        return await self._wait_for(condition=predicate, truthiness_to_wait_for=False)

    async def _wait_for(
        self, condition: Callable[[], _ReturnT], truthiness_to_wait_for: bool
    ) -> _ReturnT:
        current_value = condition()

        while bool(current_value) != truthiness_to_wait_for:
            await self._change_notifier.wait()
            current_value = condition()

        return current_value

    def _get_next_state(self) -> State:
        """Get a new instance of the state value object."""
        return State(
            commands=self._command_store.state,
            addressable_areas=self._addressable_area_store.state,
            labware=self._labware_store.state,
            pipettes=self._pipette_store.state,
            modules=self._module_store.state,
            liquids=self._liquid_store.state,
            tips=self._tip_store.state,
            wells=self._well_store.state,
            files=self._file_store.state,
        )

    def _initialize_state(self) -> None:
        """Initialize state data and view."""
        state = self._get_next_state()

        # Base states
        self._state = state
        self._commands = CommandView(state.commands)
        self._addressable_areas = AddressableAreaView(state.addressable_areas)
        self._labware = LabwareView(state.labware)
        self._pipettes = PipetteView(state.pipettes)
        self._modules = ModuleView(state.modules)
        self._liquid = LiquidView(state.liquids)
        self._tips = TipView(state.tips)
        self._wells = WellView(state.wells)
        self._files = FileView(state.files)

        # Derived states
        self._geometry = GeometryView(
            config=self._config,
            labware_view=self._labware,
            well_view=self._wells,
            module_view=self._modules,
            pipette_view=self._pipettes,
            addressable_area_view=self._addressable_areas,
        )
        self._motion = MotionView(
            config=self._config,
            labware_view=self._labware,
            pipette_view=self._pipettes,
            addressable_area_view=self._addressable_areas,
            geometry_view=self._geometry,
            module_view=self._modules,
        )

    def _update_state_views(self) -> None:
        """Update state view interfaces to use latest underlying values."""
        next_state = self._get_next_state()
        self._state = next_state
        self._commands._state = next_state.commands
        self._addressable_areas._state = next_state.addressable_areas
        self._labware._state = next_state.labware
        self._pipettes._state = next_state.pipettes
        self._modules._state = next_state.modules
        self._liquid._state = next_state.liquids
        self._tips._state = next_state.tips
        self._wells._state = next_state.wells
        self._change_notifier.notify()
        if self._notify_robot_server is not None:
            self._notify_robot_server()



"""Dispense-in-place command request, result, and implementation models."""
from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Union
from typing_extensions import Literal
from pydantic import Field

from opentrons_shared_data.errors.exceptions import PipetteOverpressureError

from .pipetting_common import (
    PipetteIdMixin,
    DispenseVolumeMixin,
    FlowRateMixin,
    BaseLiquidHandlingResult,
    OverpressureError,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    SuccessData,
    DefinedErrorData,
)
from ..errors.error_occurrence import ErrorOccurrence
from ..state.update_types import StateUpdate, CLEAR
from ..types import CurrentWell

if TYPE_CHECKING:
    from ..execution import PipettingHandler, GantryMover
    from ..resources import ModelUtils
    from ..state.state import StateView


DispenseInPlaceCommandType = Literal["dispenseInPlace"]


class DispenseInPlaceParams(PipetteIdMixin, DispenseVolumeMixin, FlowRateMixin):
    """Payload required to dispense in place."""

    pushOut: Optional[float] = Field(
        None,
        description="push the plunger a small amount farther than necessary for accurate low-volume dispensing",
    )


class DispenseInPlaceResult(BaseLiquidHandlingResult):
    """Result data from the execution of a DispenseInPlace command."""

    pass


_ExecuteReturn = Union[
    SuccessData[DispenseInPlaceResult],
    DefinedErrorData[OverpressureError],
]


class DispenseInPlaceImplementation(
    AbstractCommandImpl[DispenseInPlaceParams, _ExecuteReturn]
):
    """DispenseInPlace command implementation."""

    def __init__(
        self,
        pipetting: PipettingHandler,
        state_view: StateView,
        gantry_mover: GantryMover,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._pipetting = pipetting
        self._state_view = state_view
        self._gantry_mover = gantry_mover
        self._model_utils = model_utils

    async def execute(self, params: DispenseInPlaceParams) -> _ExecuteReturn:
        """Dispense without moving the pipette."""
        state_update = StateUpdate()
        current_location = self._state_view.pipettes.get_current_location()
        current_position = await self._gantry_mover.get_position(params.pipetteId)
        try:
            volume = await self._pipetting.dispense_in_place(
                pipette_id=params.pipetteId,
                volume=params.volume,
                flow_rate=params.flowRate,
                push_out=params.pushOut,
            )
        except PipetteOverpressureError as e:
            if (
                isinstance(current_location, CurrentWell)
                and current_location.pipette_id == params.pipetteId
            ):
                state_update.set_liquid_operated(
                    labware_id=current_location.labware_id,
                    well_name=current_location.well_name,
                    volume_added=CLEAR,
                )
            return DefinedErrorData(
                public=OverpressureError(
                    id=self._model_utils.generate_id(),
                    createdAt=self._model_utils.get_timestamp(),
                    wrappedErrors=[
                        ErrorOccurrence.from_failed(
                            id=self._model_utils.generate_id(),
                            createdAt=self._model_utils.get_timestamp(),
                            error=e,
                        )
                    ],
                    errorInfo=(
                        {
                            "retryLocation": (
                                current_position.x,
                                current_position.y,
                                current_position.z,
                            )
                        }
                    ),
                ),
                state_update=state_update,
            )
        else:
            if (
                isinstance(current_location, CurrentWell)
                and current_location.pipette_id == params.pipetteId
            ):
                state_update.set_liquid_operated(
                    labware_id=current_location.labware_id,
                    well_name=current_location.well_name,
                    volume_added=volume,
                )
            return SuccessData(
                public=DispenseInPlaceResult(volume=volume),
                state_update=state_update,
            )


class DispenseInPlace(
    BaseCommand[DispenseInPlaceParams, DispenseInPlaceResult, OverpressureError]
):
    """DispenseInPlace command model."""

    commandType: DispenseInPlaceCommandType = "dispenseInPlace"
    params: DispenseInPlaceParams
    result: Optional[DispenseInPlaceResult]

    _ImplementationCls: Type[
        DispenseInPlaceImplementation
    ] = DispenseInPlaceImplementation


class DispenseInPlaceCreate(BaseCommandCreate[DispenseInPlaceParams]):
    """DispenseInPlace command request model."""

    commandType: DispenseInPlaceCommandType = "dispenseInPlace"
    params: DispenseInPlaceParams

    _CommandCls: Type[DispenseInPlace] = DispenseInPlace



"""Blow-out in place command request, result, and implementation models."""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Union
from opentrons_shared_data.errors.exceptions import PipetteOverpressureError
from typing_extensions import Literal
from pydantic import BaseModel

from .pipetting_common import (
    OverpressureError,
    PipetteIdMixin,
    FlowRateMixin,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)
from ..errors.error_occurrence import ErrorOccurrence

from opentrons.hardware_control import HardwareControlAPI


if TYPE_CHECKING:
    from ..execution import PipettingHandler, GantryMover
    from ..state.state import StateView
    from ..resources import ModelUtils


BlowOutInPlaceCommandType = Literal["blowOutInPlace"]


class BlowOutInPlaceParams(PipetteIdMixin, FlowRateMixin):
    """Payload required to blow-out in place."""

    pass


class BlowOutInPlaceResult(BaseModel):
    """Result data from the execution of a BlowOutInPlace command."""

    pass


_ExecuteReturn = Union[
    SuccessData[BlowOutInPlaceResult],
    DefinedErrorData[OverpressureError],
]


class BlowOutInPlaceImplementation(
    AbstractCommandImpl[BlowOutInPlaceParams, _ExecuteReturn]
):
    """BlowOutInPlace command implementation."""

    def __init__(
        self,
        pipetting: PipettingHandler,
        state_view: StateView,
        hardware_api: HardwareControlAPI,
        model_utils: ModelUtils,
        gantry_mover: GantryMover,
        **kwargs: object,
    ) -> None:
        self._pipetting = pipetting
        self._state_view = state_view
        self._hardware_api = hardware_api
        self._model_utils = model_utils
        self._gantry_mover = gantry_mover

    async def execute(self, params: BlowOutInPlaceParams) -> _ExecuteReturn:
        """Blow-out without moving the pipette."""
        try:
            current_position = await self._gantry_mover.get_position(params.pipetteId)
            await self._pipetting.blow_out_in_place(
                pipette_id=params.pipetteId, flow_rate=params.flowRate
            )
        except PipetteOverpressureError as e:
            return DefinedErrorData(
                public=OverpressureError(
                    id=self._model_utils.generate_id(),
                    createdAt=self._model_utils.get_timestamp(),
                    wrappedErrors=[
                        ErrorOccurrence.from_failed(
                            id=self._model_utils.generate_id(),
                            createdAt=self._model_utils.get_timestamp(),
                            error=e,
                        )
                    ],
                    errorInfo={
                        "retryLocation": (
                            current_position.x,
                            current_position.y,
                            current_position.z,
                        )
                    },
                ),
            )
        else:
            return SuccessData(
                public=BlowOutInPlaceResult(),
            )


class BlowOutInPlace(
    BaseCommand[BlowOutInPlaceParams, BlowOutInPlaceResult, ErrorOccurrence]
):
    """BlowOutInPlace command model."""

    commandType: BlowOutInPlaceCommandType = "blowOutInPlace"
    params: BlowOutInPlaceParams
    result: Optional[BlowOutInPlaceResult]

    _ImplementationCls: Type[
        BlowOutInPlaceImplementation
    ] = BlowOutInPlaceImplementation


class BlowOutInPlaceCreate(BaseCommandCreate[BlowOutInPlaceParams]):
    """BlowOutInPlace command request model."""

    commandType: BlowOutInPlaceCommandType = "blowOutInPlace"
    params: BlowOutInPlaceParams

    _CommandCls: Type[BlowOutInPlace] = BlowOutInPlace



"""A helper for controlling a `ProtocolEngine` without async/await."""
from asyncio import AbstractEventLoop, run_coroutine_threadsafe
from typing import Any, Final, overload
from typing_extensions import Literal

from opentrons_shared_data.labware.types import LabwareUri
from opentrons_shared_data.labware.labware_definition import LabwareDefinition


from ..protocol_engine import ProtocolEngine
from ..errors import ProtocolCommandFailedError
from ..error_recovery_policy import ErrorRecoveryType
from ..state.state import StateView
from ..commands import Command, CommandCreate, CommandResult, CommandStatus


class RunStoppedBeforeCommandError(RuntimeError):
    """Raised if the ProtocolEngine was stopped before a command could start."""

    def __init__(self, command: Command) -> None:
        self._command = command
        super().__init__(
            f"The run was stopped"
            f" before {command.commandType} command {command.id} could execute."
        )


class ChildThreadTransport:
    """A helper for controlling a `ProtocolEngine` without async/await.

    You shouldn't use this directly, except to construct it and then pass it to a `SyncClient`.

    This class is responsible for doing the actual transformation from async `ProtocolEngine` calls
    to non-async ones, and doing it in a thread-safe way.
    """

    def __init__(self, engine: ProtocolEngine, loop: AbstractEventLoop) -> None:
        """Initialize the `ChildThreadTransport`.

        Args:
            engine: The `ProtocolEngine` instance that you want to interact with.
                It must be running in a thread *other* than the one from which you
                want to synchronously access it.
            loop: The event loop that `engine` is running in (in the other thread).
        """
        # We might access these from different threads,
        # so let's make them Final for (shallow) immutability.
        self._engine: Final = engine
        self._loop: Final = loop

    @property
    def state(self) -> StateView:
        """Get a view of the Protocol Engine's state."""
        return self._engine.state_view

    def execute_command(self, request: CommandCreate) -> CommandResult:
        """Execute a ProtocolEngine command.

        This blocks until the command completes. If the command fails, this will always
        raise the failure as an exception--even if ProtocolEngine deemed the failure
        recoverable.

        Args:
            request: The ProtocolEngine command request

        Returns:
            The command's result data.

        Raises:
            ProtocolEngineError: If the command execution was not successful,
                the specific error that caused the command to fail is raised.

                If the run was stopped before the command could complete, that's
                also signaled as this exception.
        """
        command = run_coroutine_threadsafe(
            self._engine.add_and_execute_command(request=request),
            loop=self._loop,
        ).result()

        # TODO: this needs to have an actual code
        if command.error is not None:
            error = command.error
            raise ProtocolCommandFailedError(
                original_error=error,
                message=f"{error.errorType}: {error.detail}",
            )

        if command.result is None:
            # This can happen with a certain pause timing:
            #
            # 1. The engine is paused.
            # 2. The user's Python script calls this method to start a new command,
            #    which remains `queued` because of the pause.
            # 3. The engine is stopped. The returned command will be `queued`
            #    and won't have a result.
            raise RunStoppedBeforeCommandError(command)

        return command.result

    def execute_command_wait_for_recovery(self, request: CommandCreate) -> Command:
        """Execute a ProtocolEngine command, including error recovery.

        This blocks until the command completes. Additionally, if the command fails,
        this will continue to block until its error recovery has been completed.

        Args:
            request: The ProtocolEngine command request.

        Returns:
            The command. If error recovery happened for it, the command will be
            reported here as failed.

        Raises:
            ProtocolEngineError: If the command failed, *and* the failure was not
                recovered from.

                If the run was stopped before the command could complete, that's
                also signalled as this exception.
        """

        async def run_in_pe_thread() -> Command:
            command = await self._engine.add_and_execute_command_wait_for_recovery(
                request=request
            )

            if command.error is not None:
                error_recovery_type = (
                    self._engine.state_view.commands.get_error_recovery_type(command.id)
                )
                error_should_fail_run = (
                    error_recovery_type == ErrorRecoveryType.FAIL_RUN
                )
                if error_should_fail_run:
                    error = command.error
                    # TODO: this needs to have an actual code
                    raise ProtocolCommandFailedError(
                        original_error=error,
                        message=f"{error.errorType}: {error.detail}",
                    )

            elif command.status == CommandStatus.QUEUED:
                # This can happen with a certain pause timing:
                #
                # 1. The engine is paused.
                # 2. The user's Python script calls this method to start a new command,
                #    which remains `queued` because of the pause.
                # 3. The engine is stopped. The returned command will be `queued`,
                #    and won't have a result.
                raise RunStoppedBeforeCommandError(command)

            return command

        command = run_coroutine_threadsafe(
            run_in_pe_thread(),
            loop=self._loop,
        ).result()

        return command

    @overload
    def call_method(
        self,
        method_name: Literal["add_labware_definition"],
        *,
        definition: LabwareDefinition,
    ) -> LabwareUri:
        ...

    @overload
    def call_method(
        self,
        method_name: Literal["reset_tips"],
        *,
        labware_id: str,
    ) -> None:
        ...

    @overload
    def call_method(
        self,
        method_name: str,
        **kwargs: Any,
    ) -> Any:
        ...

    def call_method(self, method_name: str, **kwargs: Any) -> Any:
        """Execute a ProtocolEngine method, returning the result."""
        return run_coroutine_threadsafe(
            self._call_method(method_name, **kwargs),
            loop=self._loop,
        ).result()

    async def _call_method(self, method_name: str, **kwargs: Any) -> Any:
        method = getattr(self._engine, method_name)
        assert callable(method), f"{method_name} is not a method of ProtocolEngine"
        return method(**kwargs)



"""opentrons.hardware_control.estop_state: module to manage estop state machine on OT-3."""

from typing import List, Optional

from opentrons_hardware.hardware_control.estop.detector import (
    EstopSummary,
    EstopDetector,
)

from opentrons.hardware_control.types import (
    EstopState,
    EstopPhysicalStatus,
    EstopAttachLocation,
    EstopStateNotification,
    HardwareEventHandler,
    HardwareEventUnsubscriber,
)


class EstopStateMachine:
    """Class to manage global Estop state."""

    def __init__(self, detector: Optional[EstopDetector]) -> None:
        """Create a new EstopStateMachine.

        If detector is None, the state machine will be initialized in
        a happy state (Disengaged, both estops detected) until it is
        hooked up to a valid detector.
        """
        self._detector: Optional[EstopDetector] = None
        self._state: EstopState = EstopState.DISENGAGED
        # Start off in a happy state until a detector is added
        self._summary = EstopSummary(
            left_detected=True, right_detected=True, engaged=False
        )
        if detector is not None:
            self.subscribe_to_detector(detector=detector)
        self._listeners: List[HardwareEventHandler] = []

    def subscribe_to_detector(self, detector: EstopDetector) -> None:
        """Configure the estop state machine to listen to a detector.

        This function will also transition the state based on the current
        status of the detector."""
        if self._detector is not None:
            self._detector.remove_listener(self.detector_listener)
        self._detector = detector
        detector.add_listener(listener=self.detector_listener)
        self._handle_state_transition(new_summary=detector.status)

    def __del__(self) -> None:
        if self._detector is not None:
            self._detector.remove_listener(self.detector_listener)

    def add_listener(self, listener: HardwareEventHandler) -> HardwareEventUnsubscriber:
        """Add a hardware event listener for estop event changes."""
        if listener not in self._listeners:
            self._listeners.append(listener)
            return lambda: self.remove_listener(listener)
        return lambda: None

    def remove_listener(self, listener: HardwareEventHandler) -> None:
        """Remove an existing hardware event listener for estop detector changes."""
        if listener in self._listeners:
            self._listeners.remove(listener)

    def detector_listener(self, summary: EstopSummary) -> None:
        """Callback from the detector."""
        self._handle_state_transition(new_summary=summary)

    @staticmethod
    def _transition_from_physically_engaged(summary: EstopSummary) -> EstopState:
        if not summary.engaged:
            # Estop disengaged, move to Logically Engaged
            return EstopState.LOGICALLY_ENGAGED
        return EstopState.PHYSICALLY_ENGAGED

    @staticmethod
    def _transition_from_disengaged(summary: EstopSummary) -> EstopState:
        if summary.engaged:
            # Estop engaged, move to physically engaged
            return EstopState.PHYSICALLY_ENGAGED
        if summary.left_detected or summary.right_detected:
            # An estop is still plugged in, stay disengaged
            return EstopState.DISENGAGED
        # Everything unplugged, block all action
        return EstopState.NOT_PRESENT

    @staticmethod
    def _transition_from_not_present(summary: EstopSummary) -> EstopState:
        if summary.engaged:
            # Estop plugged in and is ON, go to physically engaged
            return EstopState.PHYSICALLY_ENGAGED
        if summary.left_detected or summary.right_detected:
            # Plugged in and NOT on, go to Disengaged
            return EstopState.DISENGAGED
        return EstopState.NOT_PRESENT

    @staticmethod
    def _transition_from_logically_engaged(summary: EstopSummary) -> EstopState:
        if summary.engaged:
            # Estop was turned on, go back to physically engaged
            return EstopState.PHYSICALLY_ENGAGED
        return EstopState.LOGICALLY_ENGAGED

    def _emit_event(self, prev_state: EstopState) -> None:
        """Broadcast a state change to all listeners."""
        event = EstopStateNotification(old_state=prev_state, new_state=self._state)
        for listener in self._listeners:
            listener(event)

    def _handle_state_transition(self, new_summary: EstopSummary) -> None:
        """Caches the new state summary and changes the _state variable."""
        self._summary = new_summary

        prev_state = self._state

        self._state = {
            EstopState.PHYSICALLY_ENGAGED: self._transition_from_physically_engaged,
            EstopState.DISENGAGED: self._transition_from_disengaged,
            EstopState.NOT_PRESENT: self._transition_from_not_present,
            EstopState.LOGICALLY_ENGAGED: self._transition_from_logically_engaged,
        }[prev_state](new_summary)

        if self._state != prev_state:
            self._emit_event(prev_state=prev_state)

    def get_physical_status(self, location: EstopAttachLocation) -> EstopPhysicalStatus:
        """Get the physical status of an attach location"""
        detected = (
            self._summary.left_detected
            if location == EstopAttachLocation.LEFT
            else self._summary.right_detected
        )

        if not detected:
            return EstopPhysicalStatus.NOT_PRESENT
        # Note that we actually *don't* have a way to check if an individual
        # estop is activated or not. But we can return Engaged or Disengaged
        # based on the global state.
        return (
            EstopPhysicalStatus.ENGAGED
            if self._summary.engaged
            else EstopPhysicalStatus.DISENGAGED
        )

    @property
    def state(self) -> EstopState:
        return self._state

    def acknowledge_and_clear(self) -> EstopState:
        """Acknowledge a `logically_engaged` status if relevant.

        If the current state is not LOGICALLY_ENGAGED, this does nothing.

        If the current state *is* LOGICALLY_ENGAGED, this will move to the
        correct return state (NOT_PRESENT or ENGAGED)"""
        if self._state == EstopState.LOGICALLY_ENGAGED:
            if self._summary.left_detected or self._summary.right_detected:
                self._state = EstopState.DISENGAGED
            else:
                self._state = EstopState.NOT_PRESENT
            self._emit_event(EstopState.LOGICALLY_ENGAGED)
        return self._state



import functools
import logging
from dataclasses import dataclass
from typing import Any, List, Optional, Tuple, Type

from opentrons import types
from opentrons.hardware_control.types import CriticalPoint
from opentrons.hardware_control.util import plan_arc

from opentrons.motion_planning import (
    DEFAULT_GENERAL_ARC_Z_MARGIN,
    DEFAULT_IN_LABWARE_ARC_Z_MARGIN,
    MINIMUM_Z_MARGIN,
    MoveType,
    get_waypoints,
)

from opentrons.protocols.api_support.labware_like import LabwareLike
from opentrons.protocol_api.core.legacy.deck import Deck
from opentrons.protocol_api.core.legacy.module_geometry import ModuleGeometry


MODULE_LOG = logging.getLogger(__name__)


class LabwareHeightError(Exception):
    pass


def max_many(*args: float) -> float:
    return functools.reduce(max, args[1:], args[0])


BAD_PAIRS = {
    ("1", "12"),
    ("12", "1"),
    ("4", "12"),
    ("12", "4"),
    ("4", "9"),
    ("9", "4"),
    ("4", "8"),
    ("8", "4"),
    ("1", "8"),
    ("8", "1"),
    ("4", "11"),
    ("11", "4"),
    ("1", "11"),
    ("11", "1"),
}


def should_dodge_thermocycler(
    deck: Deck, from_loc: types.Location, to_loc: types.Location
) -> bool:
    """
    Decide if the requested path would cross the thermocycler, if
    installed.

    Returns True if we need to dodge, False otherwise
    """
    if deck.thermocycler_present:
        transit = (from_loc.labware.first_parent(), to_loc.labware.first_parent())
        # mypy doesn't like this because transit could be none, but it's
        # checked by value in BAD_PAIRS which has only strings
        return transit in BAD_PAIRS

    return False


@dataclass
class MoveConstraints:
    instr_max_height: float
    well_z_margin: float = DEFAULT_IN_LABWARE_ARC_Z_MARGIN
    lw_z_margin: float = DEFAULT_GENERAL_ARC_Z_MARGIN
    minimum_lw_z_margin: float = MINIMUM_Z_MARGIN
    minimum_z_height: float = 0.0

    @classmethod
    def build(cls: Type["MoveConstraints"], **kwargs: Any) -> "MoveConstraints":
        return cls(**{k: v for k, v in kwargs.items() if v is not None})


def get_move_type(
    from_loc: types.Location,
    to_loc: types.Location,
    force_direct: bool = False,
) -> MoveType:
    """Given two Locations, return the type of move."""
    move_type = MoveType.GENERAL_ARC
    from_labware, from_well = from_loc.labware.get_parent_labware_and_well()
    to_labware, to_well = to_loc.labware.get_parent_labware_and_well()
    same_labware = to_labware is not None and to_labware == from_labware
    same_well = to_well is not None and to_well == from_well

    if same_labware:
        move_type = MoveType.DIRECT if same_well else MoveType.IN_LABWARE_ARC

    return move_type if not force_direct else MoveType.DIRECT


def safe_height(
    from_loc: types.Location,
    to_loc: types.Location,
    deck: Deck,
    instr_max_height: float,
    well_z_margin: Optional[float] = None,
    lw_z_margin: Optional[float] = None,
    minimum_lw_z_margin: Optional[float] = None,
    minimum_z_height: Optional[float] = None,
) -> float:
    """
    Derive the height required to clear the current deck setup along
    with other constraints
    :param from_loc: The last location.
    :param to_loc: The location to move to.
    :param deck: The :py:class:`Deck` instance describing the robot.
    :param float instr_max_height: The highest z location this pipette can
                                   achieve
    :param float well_z_margin: How much extra Z margin to raise the cp by over
                                the bare minimum to clear wells within the same
                                labware. Default: 5mm
    :param float lw_z_margin: How much extra Z margin to raise the cp by over
                              the bare minimum to clear different pieces of
                              labware.
    :param minimum_z_height: When specified, this Z margin is able to raise
                             (but never lower) the mid-arc height.
    """
    constraints = MoveConstraints.build(
        instr_max_height=instr_max_height,
        well_z_margin=well_z_margin,
        lw_z_margin=lw_z_margin,
        minimum_lw_z_margin=minimum_lw_z_margin,
        minimum_z_height=minimum_z_height,
    )
    assert constraints.minimum_z_height >= 0.0
    return _build_safe_height(from_loc, to_loc, deck, constraints)


def _build_safe_height(
    from_loc: types.Location,
    to_loc: types.Location,
    deck: Deck,
    constraints: MoveConstraints,
) -> float:
    to_point = to_loc.point
    to_lw, to_well = to_loc.labware.get_parent_labware_and_well()
    from_point = from_loc.point
    from_lw, from_well = from_loc.labware.get_parent_labware_and_well()

    if to_lw and from_lw and to_lw == from_lw:
        # If we know the labwares we’re moving from and to, we can calculate
        # a safe z based on their heights
        if to_well:
            to_safety = to_well.top().point.z + constraints.well_z_margin
        else:
            to_safety = to_lw.highest_z + constraints.well_z_margin
        if from_well:
            from_safety = from_well.top().point.z + constraints.well_z_margin
        else:
            from_safety = from_lw.highest_z + constraints.well_z_margin
        # if we are already at the labware, we know the instr max height would
        # be tall enough
        if max(from_safety, to_safety) > constraints.instr_max_height:
            to_safety = constraints.instr_max_height
            from_safety = 0.0  # (ignore since it's in a max())
    else:
        # One of our labwares is invalid so we have to just go above
        # deck.highest_z since we don’t know where we are
        to_safety = deck.highest_z + constraints.lw_z_margin

        if to_safety > constraints.instr_max_height:
            if constraints.instr_max_height >= (
                deck.highest_z + constraints.minimum_lw_z_margin
            ):
                to_safety = constraints.instr_max_height
            else:
                tallest_lw = list(
                    filter(
                        lambda lw: lw.highest_z == deck.highest_z,
                        [lw for lw in deck.data.values() if lw],
                    )
                )[0]
                if isinstance(tallest_lw, ModuleGeometry) and tallest_lw.labware:
                    tallest_lw = tallest_lw.labware
                raise LabwareHeightError(
                    f"The {tallest_lw} has a total height of {deck.highest_z}"
                    " mm, which is too tall for your current pipette "
                    "configurations. The longest pipette on your robot can "
                    f"only be raised to {constraints.instr_max_height} mm "
                    "above the deck. "
                    "This may be because the labware is incorrectly defined, "
                    "incorrectly calibrated, or physically too tall. This "
                    "could also be caused by the pipette and its tipracks "
                    "being mismatched. Please check your protocol, labware "
                    "definitions and calibrations."
                )
        from_safety = 0.0  # (ignore since it’s in a max())

    return max_many(
        to_point.z, from_point.z, to_safety, from_safety, constraints.minimum_z_height
    )


def plan_moves(
    from_loc: types.Location,
    to_loc: types.Location,
    deck: Deck,
    instr_max_height: float,
    well_z_margin: Optional[float] = None,
    lw_z_margin: Optional[float] = None,
    force_direct: bool = False,
    minimum_lw_z_margin: Optional[float] = None,
    minimum_z_height: Optional[float] = None,
    use_experimental_waypoint_planning: bool = False,
) -> List[Tuple[types.Point, Optional[CriticalPoint]]]:
    """Plan moves between one :py:class:`.Location` and another.

    Each :py:class:`.Location` instance might or might not have a specific
    kind of geometry attached. This function is intended to return series
    of moves that contain the minimum safe retractions to avoid (known)
    labware on the specified :py:class:`Deck`.
    :param from_loc: The last location.
    :param to_loc: The location to move to.
    :param deck: The :py:class:`Deck` instance describing the robot.
    :param force_direct: If True, ignore any Z margins force a direct move
    :param use_experimental_waypoint_planning: If True, use waypoint logic
                                               from opentrons.motion_planning

    The other parameters are as :py:meth:`safe_height`.

    :returns: A list of tuples of :py:class:`.Point` and critical point
              overrides to move through.
    """
    constraints = MoveConstraints.build(
        instr_max_height=instr_max_height,
        well_z_margin=well_z_margin,
        lw_z_margin=lw_z_margin,
        minimum_lw_z_margin=minimum_lw_z_margin,
        minimum_z_height=minimum_z_height,
    )
    assert constraints.minimum_z_height >= 0.0

    to_point = to_loc.point
    to_lw, to_well = to_loc.labware.get_parent_labware_and_well()
    from_point = from_loc.point
    from_lw, from_well = from_loc.labware.get_parent_labware_and_well()
    from_center = LabwareLike(from_lw).center_multichannel_on_wells()
    to_center = LabwareLike(to_lw).center_multichannel_on_wells()
    dest_cp_override = CriticalPoint.XY_CENTER if to_center else None
    origin_cp_override = CriticalPoint.XY_CENTER if from_center else None
    extra_waypoints = []

    if should_dodge_thermocycler(deck, from_loc, to_loc):
        sc = deck.get_slot_center("5")
        extra_waypoints = [(sc.x, sc.y)]

    if use_experimental_waypoint_planning:
        move_type = get_move_type(from_loc, to_loc, force_direct)
        min_travel_z = deck.highest_z

        if to_lw is not None and move_type == MoveType.IN_LABWARE_ARC:
            min_travel_z = to_lw.highest_z

        # TODO(mc, 2020-11-05): if this ever needs to be used, we need a
        # story to re-create error messaging from LabwareHeightError above if
        # this raises a MotionPlanningError
        waypoints = get_waypoints(
            origin=from_point,
            dest=to_point,
            min_travel_z=min_travel_z,
            max_travel_z=instr_max_height,
            move_type=move_type,
            xy_waypoints=extra_waypoints,
            origin_cp=origin_cp_override,
            dest_cp=dest_cp_override,
        )
        return [(wp.position, wp.critical_point) for wp in waypoints]

    is_same_location = (to_lw and to_lw == from_lw) and (
        to_well and to_well == from_well
    )
    if force_direct or (is_same_location and not (minimum_z_height or 0) > 0):
        # If we’re going direct, we can assume we’re already in the correct
        # cp so we can use the override without prep
        return [(to_point, dest_cp_override)]

    # Find the safe z heights based on the destination and origin labware/well
    safe = _build_safe_height(from_loc, to_loc, deck, constraints)

    return plan_arc(
        from_point,
        to_point,
        safe,
        origin_cp_override,
        dest_cp_override,
        extra_waypoints,
    )



# noqa: D100
from __future__ import annotations
from typing import TYPE_CHECKING

from .actions import (
    Action,
    ResumeFromRecoveryAction,
    SucceedCommandAction,
    FailCommandAction,
)
from ..commands.command import DefinedErrorData
from ..error_recovery_policy import ErrorRecoveryType

if TYPE_CHECKING:
    from ..state.update_types import StateUpdate


def get_state_updates(action: Action) -> list[StateUpdate]:
    """Extract all the StateUpdates that the StateStores should apply when they apply an action."""
    if isinstance(action, SucceedCommandAction):
        return [action.state_update]

    elif isinstance(action, FailCommandAction) and isinstance(
        action.error, DefinedErrorData
    ):
        if action.type == ErrorRecoveryType.ASSUME_FALSE_POSITIVE_AND_CONTINUE:
            return [
                action.error.state_update,
                action.error.state_update_if_false_positive,
            ]
        else:
            return [action.error.state_update]

    elif isinstance(action, ResumeFromRecoveryAction):
        return [action.state_update]

    else:
        return []



"""
Coordinate subsystem detection and updates.
"""
import asyncio
from contextlib import contextmanager, ExitStack
from dataclasses import dataclass
import logging
from typing import (
    Optional,
    Set,
    Dict,
    Iterator,
    Callable,
    AsyncIterator,
    Union,
    List,
)

from opentrons_hardware.hardware_control import network, tools
from opentrons_hardware.drivers import can_bus, binary_usb
from opentrons_hardware.firmware_bindings.constants import (
    NodeId,
    USBTarget,
    FirmwareTarget,
    ToolType,
)
from opentrons_hardware.firmware_update import FirmwareUpdate

from ..types import SubSystem, SubSystemState, UpdateStatus, UpdateState
from .ot3utils import subsystem_to_target, target_to_subsystem
from .errors import SubsystemUpdating


log = logging.getLogger(__name__)


@dataclass
class FirmwareUpdateRequirements:
    target_version: int
    filepath: str


class SubsystemManager:
    """Control tracking and updating attached subsystems to the machine.

    Call start() before trying to run the system (or build with build()).

    This object has an always-up-to-date cache of attached subsystems driven by
    asynchronous notifications from hardware and explicitly updatable with refresh().
    """

    _can_messenger: can_bus.CanMessenger
    _usb_messenger: Optional[binary_usb.BinaryMessenger]
    _tool_detector: tools.detector.ToolDetector
    _network_info: network.NetworkInfo
    _tool_detection_task: "Optional[asyncio.Task[None]]"
    _expected_core_targets: Set[FirmwareTarget]
    _present_tools: tools.types.ToolSummary
    _tool_task_condition: asyncio.Condition
    _tool_task_state: Union[bool, Exception]
    _updates_required: Dict[FirmwareTarget, FirmwareUpdateRequirements]
    _updates_ongoing: Dict[SubSystem, UpdateStatus]
    _update_bag: FirmwareUpdate

    def __init__(
        self,
        can_messenger: can_bus.CanMessenger,
        usb_messenger: Optional[binary_usb.BinaryMessenger],
        tool_detector: tools.detector.ToolDetector,
        network_info: network.NetworkInfo,
        update_bag: FirmwareUpdate,
    ) -> None:
        self._can_messenger = can_messenger
        self._usb_messenger = usb_messenger
        self._tool_detector = tool_detector
        self._network_info = network_info
        self._tool_detection_task = None
        self._expected_core_targets: Set[FirmwareTarget] = {
            NodeId.gantry_x,
            NodeId.gantry_y,
            NodeId.head,
        }
        self._tool_task_condition = asyncio.Condition()
        self._tool_task_state = False
        self._updates_required = {}
        self._updates_ongoing = {}
        self._update_bag = update_bag
        if self._usb_messenger:
            self._expected_core_targets.add(USBTarget.rear_panel)
        self._present_tools = tools.types.ToolSummary(
            left=None, right=None, gripper=None
        )

    @property
    def ok(self) -> bool:
        """True if required subsystems are present and all present subsystems are ok."""
        if not self._expected_core_targets.issubset(self.targets):
            return False
        return all(state.ok for state in self.subsystems.values())

    @property
    def device_info(self) -> Dict[SubSystem, network.DeviceInfoCache]:
        return {
            target_to_subsystem(target): info
            for target, info in self._network_info.device_info.items()
        }

    @property
    def targets(self) -> Set[FirmwareTarget]:
        return self._network_info.targets

    @property
    def tools(self) -> tools.types.ToolSummary:
        return self._present_tools

    @property
    def subsystems(self) -> Dict[SubSystem, SubSystemState]:
        def _state_or(maybe_status: Optional[UpdateStatus]) -> Optional[UpdateState]:
            if maybe_status is not None:
                return maybe_status.state
            return None

        def _next_version(target: FirmwareTarget, current: int) -> int:
            try:
                return self._updates_required[target].target_version
            except KeyError:
                return current

        return {
            target_to_subsystem(target): SubSystemState(
                ok=info.ok,
                current_fw_version=info.version,
                next_fw_version=_next_version(target, info.version),
                fw_update_needed=target in self._updates_required,
                current_fw_sha=info.shortsha,
                pcba_revision=str(info.revision),
                update_state=_state_or(
                    self._updates_ongoing.get(target_to_subsystem(target))
                ),
            )
            for target, info in self._network_info.device_info.items()
        }

    @property
    def update_required(self) -> bool:
        return bool(self._updates_required)

    @property
    def subsystems_to_update(self) -> List[SubSystem]:
        return [target_to_subsystem(t) for t in self._updates_required.keys()]

    async def start(self) -> None:
        await self._probe_network_and_cache_fw_updates(
            self._expected_core_targets, True
        )
        self._tool_detection_task = asyncio.create_task(
            self._tool_detection_task_main()
        )
        await self.refresh()
        log.info(f"Subsystem manager started with devices {self.device_info}")

    async def refresh(self) -> None:
        """Explicitly refresh the state of the system.

        This will not return until the information is up to date.
        """
        async with self._tool_task_condition:
            if isinstance(self._tool_task_state, Exception):
                raise self._tool_task_state
            await self._tool_detector.check_once()
            await self._tool_task_condition.wait()
            if isinstance(self._tool_task_state, Exception):
                raise self._tool_task_state
            self._tool_task_state = False

    async def is_new(self) -> bool:
        """Returns True if there is data unread since the last call to is_new."""
        async with self._tool_task_condition:
            state = self._tool_task_state
            if isinstance(state, Exception):
                raise state
            return self._tool_task_state is True

    async def _check_devices_after_update(
        self, devices: Set[SubSystem], timeout_sec: float = 10.0
    ) -> None:
        try:
            await asyncio.wait_for(
                self._do_check_devices_after_update(devices), timeout=timeout_sec
            )
        except asyncio.TimeoutError:
            raise RuntimeError("Device failed to come back after firmware update")

    async def _do_check_devices_after_update(self, devices: Set[SubSystem]) -> None:
        # refresh the device_info cache and reset internal states
        log.info(f"Checking {devices} after update")
        attempts = 1
        while True:
            await self._probe_network_and_cache_fw_updates(
                {subsystem_to_target(subsystem) for subsystem in devices},
                False,
            )
            await asyncio.sleep(0.1)
            # make sure the update actually succeeded
            device_info = self.device_info
            if devices.issubset(device_info.keys()) and all(
                info.ok for subsys, info in device_info.items() if subsys in devices
            ):
                log.info(f"Found all devices after {attempts} tries")
                return
            log.info(f"Failed to find all of {devices}: found {device_info}")
            attempts += 1

    def get_update_progress(self) -> Dict[SubSystem, UpdateStatus]:
        """Returns a set of UpdateStatus of the updates."""
        return {k: v for k, v in self._updates_ongoing.items()}

    async def update_firmware(
        self,
        subsystems: Optional[Set[SubSystem]] = None,
        force: bool = False,
    ) -> AsyncIterator[UpdateStatus]:
        """Updates the firmware on the OT3."""
        # Check that there arent updates already running for given nodes
        current_subsystems = self.subsystems

        subsystems_to_check = subsystems or set(current_subsystems)
        subsystems_updating = set(self._updates_ongoing)
        subsystems_to_update = subsystems_to_check - subsystems_updating
        if not subsystems_to_update:
            if subsystems:
                raise SubsystemUpdating(f"Ongoing update for {subsystems}")
            else:
                log.info("No viable subsystem to update.")
                return

        # Check if devices need an update, only checks nodes if given
        firmware_updates = self._get_required_fw_updates(
            {subsystem_to_target(subsystem) for subsystem in subsystems_to_update},
            force=force,
        )

        subsystems_updatable = set(
            target_to_subsystem(target) for target in firmware_updates.keys()
        )
        updating_subsystems = subsystems_to_update.intersection(subsystems_updatable)

        if not updating_subsystems:
            log.info("No firmware updates required for specified subsystems.")
            return

        log.info(f"Updating firmware on {updating_subsystems}")

        update_details = {
            target: firmware_updates[target].filepath
            for target in (
                subsystem_to_target(subsystem) for subsystem in updating_subsystems
            )
        }

        with ExitStack() as update_tracker_stack:
            status_callbacks: Dict[SubSystem, Callable[[UpdateStatus], None]] = {
                subsystem: update_tracker_stack.enter_context(
                    self._update_ongoing_for(subsystem)
                )
                for subsystem in updating_subsystems
            }

            updater = self._update_bag.update_runner(
                can_messenger=self._can_messenger,
                usb_messenger=self._usb_messenger,
                update_details=update_details,
                retry_count=3,
                timeout_seconds=5,
                erase=True,
            )

            # start the updates and yield progress to caller
            async for target, status_element in updater.run_updates():
                subsystem = target_to_subsystem(target)
                upstream_status = UpdateStatus(
                    subsystem=subsystem,
                    state=UpdateState[status_element[0].name],
                    progress=int(status_element[1] * 100),
                )
                status_callbacks[subsystem](upstream_status)
                yield upstream_status

        await self._check_devices_after_update(updating_subsystems)

    def _get_required_fw_updates(
        self, targets: Set[FirmwareTarget], force: bool
    ) -> Dict[FirmwareTarget, FirmwareUpdateRequirements]:
        to_check = (targets or self.targets).intersection(self.targets)
        updates_required: Dict[FirmwareTarget, FirmwareUpdateRequirements] = {}
        bad_targets = {
            target
            for target in to_check
            if not self.device_info[target_to_subsystem(target)].ok
        }
        good_targets = to_check - bad_targets
        device_info_by_target = {
            subsystem_to_target(subsystem): info
            for subsystem, info in self.device_info.items()
        }
        good_updates = self._update_bag.update_checker(
            device_info_by_target, good_targets, force
        )
        bad_updates = self._update_bag.update_checker(
            device_info_by_target, bad_targets, True
        )

        for target, (next_version, filepath) in good_updates.items():
            updates_required[target] = FirmwareUpdateRequirements(
                target_version=next_version,
                filepath=filepath,
            )

        for target, (next_version, filepath) in bad_updates.items():
            updates_required[target] = FirmwareUpdateRequirements(
                target_version=next_version,
                filepath=filepath,
            )

        return updates_required

    def _update_fw_update_requirements(self) -> None:
        """Update the local cache of extra information required for subsystem state."""
        # this runs in a separate task from where it will be queried, so we'll build a replacement
        # for the _updates_required dict locally and then swap it over async-atomically

        self._updates_required = self._get_required_fw_updates(set(), False)

    async def _probe_network_and_cache_fw_updates(
        self, targets: Set[FirmwareTarget], broadcast: bool = True
    ) -> None:
        checked_targets = {
            target
            for target in targets
            if target_to_subsystem(target) not in self._updates_ongoing
        }
        if broadcast:
            await self._network_info.probe(checked_targets)
        else:
            await self._network_info.probe_specific(checked_targets)
        self._update_fw_update_requirements()

    @contextmanager
    def _update_ongoing_for(
        self, subsystem: SubSystem
    ) -> Iterator[Callable[[UpdateStatus], None]]:
        if subsystem in self._updates_ongoing:
            raise RuntimeError(f"Update already ongoing for {subsystem}")

        def update_state(status: UpdateStatus) -> None:
            self._updates_ongoing[subsystem] = status

        self._updates_ongoing[subsystem] = UpdateStatus(
            subsystem=subsystem, state=UpdateState.queued, progress=0
        )
        try:
            yield update_state
        except Exception:
            log.exception(f"Update failed for subsystem {subsystem}")
            raise
        finally:
            self._updates_ongoing.pop(subsystem)

    async def _tool_detection_task_main(self) -> None:
        try:
            await self._tool_detection_task_protected()
        except Exception as e:
            async with self._tool_task_condition:
                self._tool_task_state = e
                self._tool_task_condition.notify_all()
            raise

    async def _tool_detection_task_protected(self) -> None:
        """Main function of an asyncio task that responds to async notifications that tool states have changed."""
        async with self._tool_task_condition:
            self._tool_task_state = True
            self._tool_task_condition.notify_all()
        async for update in self._tool_detector.detect():
            # We got a notification from the head that something changed. Let's check whether the currently
            # attached tools are okay by getting their device info
            tool_nodes: Set[NodeId] = set()
            if update.left != ToolType.nothing_attached:
                tool_nodes.add(NodeId.pipette_left)
            if update.right != ToolType.nothing_attached:
                tool_nodes.add(NodeId.pipette_right)
            if update.gripper != ToolType.nothing_attached:
                tool_nodes.add(NodeId.gripper)
            base_nodes: Set[FirmwareTarget] = {
                NodeId.pipette_left,
                NodeId.pipette_right,
                NodeId.gripper,
            }
            try:
                self._network_info.mark_absent(base_nodes - tool_nodes)
                await self._probe_network_and_cache_fw_updates(
                    self._expected_core_targets.union(tool_nodes), False
                )
            except TimeoutError:
                log.exception("Problem in internal network probe")
                continue

            # Once we know which are okay, we can ask only those for their subsystems
            to_resolve = tools.types.ToolDetectionResult(
                left=self._tool_if_ok(update.left, NodeId.pipette_left),
                right=self._tool_if_ok(update.right, NodeId.pipette_right),
                gripper=self._tool_if_ok(update.gripper, NodeId.gripper),
            )
            self._present_tools = await self._tool_detector.resolve(to_resolve, 10.0)
            log.info(f"Present tools are now {self._present_tools}")
            async with self._tool_task_condition:
                self._tool_task_state = True
                self._tool_task_condition.notify_all()

    def _tool_if_ok(self, tool: ToolType, node: NodeId) -> ToolType:
        if tool is ToolType.nothing_attached:
            return tool
        device_info = self._network_info.device_info
        if node not in device_info:
            return ToolType.nothing_attached
        if not device_info[node].ok:
            return ToolType.nothing_attached
        return tool



from typing import (
    Dict,
    Optional,
    Protocol,
    List,
    Literal,
    Tuple,
    Union,
    runtime_checkable,
    TypeVar,
)

Response = TypeVar("Response")

ErrorCodeNames = Literal[
    "NO_ERROR",
    "UNKNOWN_ERROR",
    "DEVICE_CLOSED",
    "INVALID_ARGUMENT",
    "NO_MEMORY",
    "UNSUPPORTED_OPERATION",
    "DEVICE_COMMUNICATION_FAILURE",
    "DEVICE_OPERATION_FAILED",
    "DEVICE_OPEN_PREFIX",
    "DEVICE_NOT_FOUND",
    "DEVICE_TOO_NEW",
    "DEVICE_ALREADY_OPEN",
    "FIRMWARE_UPDATE_ERROR_PREFIX",
    "FIRMWARE_UPDATE_FILE_NOT_FOUND",
    "FIRMWARE_UPDATE_FILE_NOT_VALID",
    "FIRMWARE_UPDATE_FAILED",
    "FILE_ERROR_PREFIX",
    "FILE_WRITE_ERROR",
    "MEASUTEMNT_ERROR_PREFIX",
    "MEASUTEMNT_SLOT_NOT_EMPTY",
    "NOT_INITIALIZED",
    "INTERNAL",
]

SlotStateNames = Literal[
    "UNKNOWN",
    "EMPTY",
    "OCCUPIED",
    "UNDETERMINED",
]

DeviceStateNames = Literal[
    "UNKNOWN",
    "OK",
    "BROKEN_FW",
    "ERROR",
]


@runtime_checkable
class AbsorbanceHidInterface(Protocol):
    @runtime_checkable
    class Device(Protocol):
        sn: str

    @runtime_checkable
    class ErrorCode(Protocol):
        __members__: Dict[ErrorCodeNames, int]
        name: ErrorCodeNames
        value: int

    @runtime_checkable
    class SlotState(Protocol):
        __members__: Dict[SlotStateNames, int]
        name: SlotStateNames
        value: int

    @runtime_checkable
    class SingleMeasurementConfig(Protocol):
        sample_wavelength: int
        reference_wavelength: Optional[int]

    @runtime_checkable
    class MultiMeasurementConfig(Protocol):
        sample_wavelengths: List[int]

    @runtime_checkable
    class DeviceInfo(Protocol):
        ref_no: str
        sn: str
        version: str

    @runtime_checkable
    class DeviceState(Protocol):
        __members__: Dict[DeviceStateNames, int]
        name: DeviceStateNames
        value: int

    def Abs96SingleMeasurementConfig(self) -> SingleMeasurementConfig:
        ...

    def Abs96MultipleMeasurementConfig(self) -> MultiMeasurementConfig:
        ...

    def open_device(self, device: Device) -> Tuple[ErrorCode, int]:
        ...

    def free_device(self, device_handle: int) -> Tuple[ErrorCode, bool]:
        ...

    def device_open(self, device_handle: int) -> bool:
        ...

    def get_device_information(
        self, device_handle: int
    ) -> Tuple[ErrorCode, DeviceInfo]:
        ...

    def update_device(self, device_handle: int, firmware_file_path: str) -> ErrorCode:
        ...

    def get_device_status(self, device_handle: int) -> Tuple[ErrorCode, DeviceState]:
        ...

    def get_device_uptime(self, device_handle: int) -> Tuple[ErrorCode, int]:
        ...

    def get_device_slot_status(self, device_handle: int) -> Tuple[ErrorCode, SlotState]:
        ...

    def get_device_parts_aligned(self, device_handle: int) -> Tuple[ErrorCode, bool]:
        ...

    def abs96_get_available_wavelengths(
        self, device_handle: int
    ) -> Tuple[ErrorCode, List[int]]:
        ...

    def abs96_initialize_single_measurement(
        self, device_handle: int, conf: SingleMeasurementConfig
    ) -> ErrorCode:
        ...

    def abs96_initialize_multiple_measurement(
        self, device_handle: int, conf: MultiMeasurementConfig
    ) -> ErrorCode:
        ...

    def abs96_single_measure(
        self, device_handle: int, conf: SingleMeasurementConfig
    ) -> Tuple[ErrorCode, List[float]]:
        ...

    def abs96_multiple_measure(
        self, device_handle: int, conf: MultiMeasurementConfig
    ) -> Tuple[ErrorCode, List[List[float]]]:
        ...

    def available_devices(self) -> List[Device]:
        ...


MeasurementConfig = Union[
    AbsorbanceHidInterface.SingleMeasurementConfig,
    AbsorbanceHidInterface.MultiMeasurementConfig,
]



"""Basic modules data state and store."""

from __future__ import annotations

from dataclasses import dataclass
from typing import (
    Dict,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    overload,
)
from numpy import array, dot, double as npdouble
from numpy.typing import NDArray

from opentrons.hardware_control.modules.magdeck import (
    OFFSET_TO_LABWARE_BOTTOM as MAGNETIC_MODULE_OFFSET_TO_LABWARE_BOTTOM,
)
from opentrons.hardware_control.modules.types import LiveData
from opentrons.motion_planning.adjacent_slots_getters import (
    get_east_slot,
    get_west_slot,
    get_adjacent_staging_slot,
)
from opentrons.protocol_engine.actions.get_state_update import get_state_updates
from opentrons.protocol_engine.commands.calibration.calibrate_module import (
    CalibrateModuleResult,
)
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.module_substates.absorbance_reader_substate import (
    AbsorbanceReaderMeasureMode,
)
from opentrons.types import DeckSlotName, MountType, StagingSlotName
from ..errors import ModuleNotConnectedError

from ..types import (
    LoadedModule,
    ModuleModel,
    ModuleOffsetVector,
    ModuleOffsetData,
    ModuleType,
    ModuleDefinition,
    DeckSlotLocation,
    ModuleDimensions,
    LabwareOffsetVector,
    HeaterShakerLatchStatus,
    HeaterShakerMovementRestrictors,
    DeckType,
    LabwareMovementOffsetData,
    AddressableAreaLocation,
)

from ..resources import DeckFixedLabware
from .addressable_areas import AddressableAreaView
from .. import errors
from ..commands import (
    Command,
    LoadModuleResult,
    heater_shaker,
    temperature_module,
    thermocycler,
    absorbance_reader,
)
from ..actions import (
    Action,
    SucceedCommandAction,
    AddModuleAction,
)
from ._abstract_store import HasState, HandlesActions
from .module_substates import (
    MagneticModuleSubState,
    HeaterShakerModuleSubState,
    TemperatureModuleSubState,
    ThermocyclerModuleSubState,
    AbsorbanceReaderSubState,
    MagneticModuleId,
    HeaterShakerModuleId,
    TemperatureModuleId,
    ThermocyclerModuleId,
    AbsorbanceReaderId,
    MagneticBlockSubState,
    MagneticBlockId,
    ModuleSubStateType,
)
from .config import Config


ModuleSubStateT = TypeVar("ModuleSubStateT", bound=ModuleSubStateType)


class SlotTransit(NamedTuple):
    """Class defining starting and ending slots in a pipette movement."""

    start: DeckSlotName
    end: DeckSlotName


_OT2_THERMOCYCLER_SLOT_TRANSITS_TO_DODGE = {
    SlotTransit(start=DeckSlotName.SLOT_1, end=DeckSlotName.FIXED_TRASH),
    SlotTransit(start=DeckSlotName.FIXED_TRASH, end=DeckSlotName.SLOT_1),
    SlotTransit(start=DeckSlotName.SLOT_4, end=DeckSlotName.FIXED_TRASH),
    SlotTransit(start=DeckSlotName.FIXED_TRASH, end=DeckSlotName.SLOT_4),
    SlotTransit(start=DeckSlotName.SLOT_4, end=DeckSlotName.SLOT_9),
    SlotTransit(start=DeckSlotName.SLOT_9, end=DeckSlotName.SLOT_4),
    SlotTransit(start=DeckSlotName.SLOT_4, end=DeckSlotName.SLOT_8),
    SlotTransit(start=DeckSlotName.SLOT_8, end=DeckSlotName.SLOT_4),
    SlotTransit(start=DeckSlotName.SLOT_1, end=DeckSlotName.SLOT_8),
    SlotTransit(start=DeckSlotName.SLOT_8, end=DeckSlotName.SLOT_1),
    SlotTransit(start=DeckSlotName.SLOT_4, end=DeckSlotName.SLOT_11),
    SlotTransit(start=DeckSlotName.SLOT_11, end=DeckSlotName.SLOT_4),
    SlotTransit(start=DeckSlotName.SLOT_1, end=DeckSlotName.SLOT_11),
    SlotTransit(start=DeckSlotName.SLOT_11, end=DeckSlotName.SLOT_1),
}

_OT3_THERMOCYCLER_SLOT_TRANSITS_TO_DODGE = {
    SlotTransit(start=t.start.to_ot3_equivalent(), end=t.end.to_ot3_equivalent())
    for t in _OT2_THERMOCYCLER_SLOT_TRANSITS_TO_DODGE
}

_THERMOCYCLER_SLOT_TRANSITS_TO_DODGE = (
    _OT2_THERMOCYCLER_SLOT_TRANSITS_TO_DODGE | _OT3_THERMOCYCLER_SLOT_TRANSITS_TO_DODGE
)

_THERMOCYCLER_SLOT = DeckSlotName.SLOT_B1
_OT2_THERMOCYCLER_ADDITIONAL_SLOTS = [
    DeckSlotName.SLOT_8,
    DeckSlotName.SLOT_10,
    DeckSlotName.SLOT_11,
]
_OT3_THERMOCYCLER_ADDITIONAL_SLOTS = [DeckSlotName.SLOT_A1]


@dataclass(frozen=True)
class HardwareModule:
    """Data describing an actually connected module."""

    serial_number: Optional[str]
    definition: ModuleDefinition


@dataclass
class ModuleState:
    """The internal data to keep track of loaded modules."""

    slot_by_module_id: Dict[str, Optional[DeckSlotName]]
    """The deck slot that each module has been loaded into.

    This will be None when the module was added via
    ProtocolEngine.use_attached_modules() instead of an explicit loadModule command.
    """

    additional_slots_occupied_by_module_id: Dict[str, List[DeckSlotName]]
    """List of additional slots occupied by each module.

    The thermocycler (both GENs), occupies multiple slots on both OT-2 and the Flex
    but only one slot is associated with the location of the thermocycler.
    In order to check for deck conflicts with other items, we will keep track of any
    additional slots occupied by a module here.

    This will be None when a module occupies only one slot.
    """

    requested_model_by_id: Dict[str, Optional[ModuleModel]]
    """The model by which each loaded module was requested.

    Becuse of module compatibility, this can differ from the model found through
    hardware_module_by_id. See `ModuleView.get_requested_model()` versus
    `ModuleView.get_connected_model()`.

    This will be None when the module was added via
    ProtocolEngine.use_attached_modules() instead of an explicit loadModule command.
    """

    hardware_by_module_id: Dict[str, HardwareModule]
    """Information about each module's physical hardware."""

    substate_by_module_id: Dict[str, ModuleSubStateType]
    """Information about each module that's specific to the module type."""

    module_offset_by_serial: Dict[str, ModuleOffsetData]
    """Information about each modules offsets."""

    deck_type: DeckType
    """Type of deck that the modules are on."""

    deck_fixed_labware: Sequence[DeckFixedLabware]
    """Fixed labware from the deck which may be assigned to a module.

    The Opentrons Plate Reader module makes use of an electronic Lid labware which moves
    between the Reader and Dock positions, and is pre-loaded into the engine as to persist
    even when not in use. For this reason, we inject it here when an appropriate match
    is identified.
    """


class ModuleStore(HasState[ModuleState], HandlesActions):
    """Module state container."""

    _state: ModuleState

    def __init__(
        self,
        config: Config,
        deck_fixed_labware: Sequence[DeckFixedLabware],
        module_calibration_offsets: Optional[Dict[str, ModuleOffsetData]] = None,
    ) -> None:
        """Initialize a ModuleStore and its state."""
        self._state = ModuleState(
            slot_by_module_id={},
            additional_slots_occupied_by_module_id={},
            requested_model_by_id={},
            hardware_by_module_id={},
            substate_by_module_id={},
            module_offset_by_serial=module_calibration_offsets or {},
            deck_type=config.deck_type,
            deck_fixed_labware=deck_fixed_labware,
        )
        self._robot_type = config.robot_type

    def handle_action(self, action: Action) -> None:
        """Modify state in reaction to an action."""
        if isinstance(action, SucceedCommandAction):
            self._handle_command(action.command)

        elif isinstance(action, AddModuleAction):
            self._add_module_substate(
                module_id=action.module_id,
                serial_number=action.serial_number,
                definition=action.definition,
                slot_name=None,
                requested_model=None,
                module_live_data=action.module_live_data,
            )

        for state_update in get_state_updates(action):
            self._handle_state_update(state_update)

    def _handle_command(self, command: Command) -> None:
        # todo(mm, 2024-11-04): Delete this function. Port these isinstance()
        # checks to the update_types.StateUpdate mechanism.

        if isinstance(command.result, LoadModuleResult):
            slot_name = command.params.location.slotName
            self._add_module_substate(
                module_id=command.result.moduleId,
                serial_number=command.result.serialNumber,
                definition=command.result.definition,
                slot_name=slot_name,
                requested_model=command.params.model,
                module_live_data=None,
            )

        if isinstance(command.result, CalibrateModuleResult):
            self._update_module_calibration(
                module_id=command.params.moduleId,
                module_offset=command.result.moduleOffset,
                location=command.result.location,
            )

        if isinstance(
            command.result,
            (
                heater_shaker.SetTargetTemperatureResult,
                heater_shaker.DeactivateHeaterResult,
                heater_shaker.SetAndWaitForShakeSpeedResult,
                heater_shaker.DeactivateShakerResult,
                heater_shaker.OpenLabwareLatchResult,
                heater_shaker.CloseLabwareLatchResult,
            ),
        ):
            self._handle_heater_shaker_commands(command)

        if isinstance(
            command.result,
            (
                temperature_module.SetTargetTemperatureResult,
                temperature_module.DeactivateTemperatureResult,
            ),
        ):
            self._handle_temperature_module_commands(command)

        if isinstance(
            command.result,
            (
                thermocycler.SetTargetBlockTemperatureResult,
                thermocycler.DeactivateBlockResult,
                thermocycler.SetTargetLidTemperatureResult,
                thermocycler.DeactivateLidResult,
                thermocycler.OpenLidResult,
                thermocycler.CloseLidResult,
            ),
        ):
            self._handle_thermocycler_module_commands(command)

        if isinstance(
            command.result,
            (
                absorbance_reader.InitializeResult,
                absorbance_reader.ReadAbsorbanceResult,
            ),
        ):
            self._handle_absorbance_reader_commands(command)

    def _handle_state_update(self, state_update: update_types.StateUpdate) -> None:
        if state_update.absorbance_reader_lid != update_types.NO_CHANGE:
            module_id = state_update.absorbance_reader_lid.module_id
            is_lid_on = state_update.absorbance_reader_lid.is_lid_on

            # Get current values:
            absorbance_reader_substate = self._state.substate_by_module_id[module_id]
            assert isinstance(
                absorbance_reader_substate, AbsorbanceReaderSubState
            ), f"{module_id} is not an absorbance plate reader."
            configured = absorbance_reader_substate.configured
            measure_mode = absorbance_reader_substate.measure_mode
            configured_wavelengths = absorbance_reader_substate.configured_wavelengths
            reference_wavelength = absorbance_reader_substate.reference_wavelength
            data = absorbance_reader_substate.data

            self._state.substate_by_module_id[module_id] = AbsorbanceReaderSubState(
                module_id=AbsorbanceReaderId(module_id),
                configured=configured,
                measured=True,
                is_lid_on=is_lid_on,
                measure_mode=measure_mode,
                configured_wavelengths=configured_wavelengths,
                reference_wavelength=reference_wavelength,
                data=data,
            )

    def _add_module_substate(
        self,
        module_id: str,
        serial_number: Optional[str],
        definition: ModuleDefinition,
        slot_name: Optional[DeckSlotName],
        requested_model: Optional[ModuleModel],
        module_live_data: Optional[LiveData],
    ) -> None:
        actual_model = definition.model
        live_data = module_live_data["data"] if module_live_data else None

        self._state.requested_model_by_id[module_id] = requested_model
        self._state.slot_by_module_id[module_id] = slot_name
        self._state.hardware_by_module_id[module_id] = HardwareModule(
            serial_number=serial_number,
            definition=definition,
        )

        if ModuleModel.is_magnetic_module_model(actual_model):
            self._state.substate_by_module_id[module_id] = MagneticModuleSubState(
                module_id=MagneticModuleId(module_id),
                model=actual_model,
            )
        elif ModuleModel.is_heater_shaker_module_model(actual_model):
            if live_data is None:
                labware_latch_status = HeaterShakerLatchStatus.UNKNOWN
            elif live_data["labwareLatchStatus"] == "idle_closed":
                labware_latch_status = HeaterShakerLatchStatus.CLOSED
            else:
                labware_latch_status = HeaterShakerLatchStatus.OPEN
            self._state.substate_by_module_id[module_id] = HeaterShakerModuleSubState(
                module_id=HeaterShakerModuleId(module_id),
                labware_latch_status=labware_latch_status,
                is_plate_shaking=(
                    live_data is not None and live_data["targetSpeed"] is not None
                ),
                plate_target_temperature=live_data["targetTemp"] if live_data else None,  # type: ignore[arg-type]
            )
        elif ModuleModel.is_temperature_module_model(actual_model):
            self._state.substate_by_module_id[module_id] = TemperatureModuleSubState(
                module_id=TemperatureModuleId(module_id),
                plate_target_temperature=live_data["targetTemp"] if live_data else None,  # type: ignore[arg-type]
            )
        elif ModuleModel.is_thermocycler_module_model(actual_model):
            self._state.substate_by_module_id[module_id] = ThermocyclerModuleSubState(
                module_id=ThermocyclerModuleId(module_id),
                is_lid_open=live_data is not None and live_data["lid"] == "open",
                target_block_temperature=live_data["targetTemp"] if live_data else None,  # type: ignore[arg-type]
                target_lid_temperature=live_data["lidTarget"] if live_data else None,  # type: ignore[arg-type]
            )
            self._update_additional_slots_occupied_by_thermocycler(
                module_id=module_id, slot_name=slot_name
            )
        elif ModuleModel.is_magnetic_block(actual_model):
            self._state.substate_by_module_id[module_id] = MagneticBlockSubState(
                module_id=MagneticBlockId(module_id)
            )
        elif ModuleModel.is_absorbance_reader(actual_model):
            self._state.substate_by_module_id[module_id] = AbsorbanceReaderSubState(
                module_id=AbsorbanceReaderId(module_id),
                configured=False,
                measured=False,
                is_lid_on=True,
                data=None,
                measure_mode=None,
                configured_wavelengths=None,
                reference_wavelength=None,
            )

    def _update_additional_slots_occupied_by_thermocycler(
        self,
        module_id: str,
        slot_name: Optional[
            DeckSlotName
        ],  # addModuleAction will not have a slot location
    ) -> None:
        if slot_name != _THERMOCYCLER_SLOT.to_equivalent_for_robot_type(
            self._robot_type
        ):
            return

        self._state.additional_slots_occupied_by_module_id[module_id] = (
            _OT3_THERMOCYCLER_ADDITIONAL_SLOTS
            if self._state.deck_type == DeckType.OT3_STANDARD
            else _OT2_THERMOCYCLER_ADDITIONAL_SLOTS
        )

    def _update_module_calibration(
        self,
        module_id: str,
        module_offset: ModuleOffsetVector,
        location: DeckSlotLocation,
    ) -> None:
        module = self._state.hardware_by_module_id.get(module_id)
        if module:
            module_serial = module.serial_number
            assert (
                module_serial is not None
            ), "Expected a module SN and got None instead."
            self._state.module_offset_by_serial[module_serial] = ModuleOffsetData(
                moduleOffsetVector=module_offset,
                location=location,
            )

    def _handle_heater_shaker_commands(
        self,
        command: Union[
            heater_shaker.SetTargetTemperature,
            heater_shaker.DeactivateHeater,
            heater_shaker.SetAndWaitForShakeSpeed,
            heater_shaker.DeactivateShaker,
            heater_shaker.OpenLabwareLatch,
            heater_shaker.CloseLabwareLatch,
        ],
    ) -> None:
        module_id = command.params.moduleId
        hs_substate = self._state.substate_by_module_id[module_id]
        assert isinstance(
            hs_substate, HeaterShakerModuleSubState
        ), f"{module_id} is not heater-shaker."

        # Get current values to preserve target temperature not being set/deactivated
        prev_state: HeaterShakerModuleSubState = hs_substate

        if isinstance(command.result, heater_shaker.SetTargetTemperatureResult):
            self._state.substate_by_module_id[module_id] = HeaterShakerModuleSubState(
                module_id=HeaterShakerModuleId(module_id),
                labware_latch_status=prev_state.labware_latch_status,
                is_plate_shaking=prev_state.is_plate_shaking,
                plate_target_temperature=command.params.celsius,
            )
        elif isinstance(command.result, heater_shaker.DeactivateHeaterResult):
            self._state.substate_by_module_id[module_id] = HeaterShakerModuleSubState(
                module_id=HeaterShakerModuleId(module_id),
                labware_latch_status=prev_state.labware_latch_status,
                is_plate_shaking=prev_state.is_plate_shaking,
                plate_target_temperature=None,
            )
        elif isinstance(command.result, heater_shaker.SetAndWaitForShakeSpeedResult):
            self._state.substate_by_module_id[module_id] = HeaterShakerModuleSubState(
                module_id=HeaterShakerModuleId(module_id),
                labware_latch_status=prev_state.labware_latch_status,
                is_plate_shaking=True,
                plate_target_temperature=prev_state.plate_target_temperature,
            )
        elif isinstance(command.result, heater_shaker.DeactivateShakerResult):
            self._state.substate_by_module_id[module_id] = HeaterShakerModuleSubState(
                module_id=HeaterShakerModuleId(module_id),
                labware_latch_status=prev_state.labware_latch_status,
                is_plate_shaking=False,
                plate_target_temperature=prev_state.plate_target_temperature,
            )
        elif isinstance(command.result, heater_shaker.OpenLabwareLatchResult):
            self._state.substate_by_module_id[module_id] = HeaterShakerModuleSubState(
                module_id=HeaterShakerModuleId(module_id),
                labware_latch_status=HeaterShakerLatchStatus.OPEN,
                is_plate_shaking=prev_state.is_plate_shaking,
                plate_target_temperature=prev_state.plate_target_temperature,
            )
        elif isinstance(command.result, heater_shaker.CloseLabwareLatchResult):
            self._state.substate_by_module_id[module_id] = HeaterShakerModuleSubState(
                module_id=HeaterShakerModuleId(module_id),
                labware_latch_status=HeaterShakerLatchStatus.CLOSED,
                is_plate_shaking=prev_state.is_plate_shaking,
                plate_target_temperature=prev_state.plate_target_temperature,
            )

    def _handle_temperature_module_commands(
        self,
        command: Union[
            temperature_module.SetTargetTemperature,
            temperature_module.DeactivateTemperature,
        ],
    ) -> None:
        module_id = command.params.moduleId
        assert isinstance(
            self._state.substate_by_module_id[module_id], TemperatureModuleSubState
        ), f"{module_id} is not a temperature module."

        if isinstance(command.result, temperature_module.SetTargetTemperatureResult):
            self._state.substate_by_module_id[module_id] = TemperatureModuleSubState(
                module_id=TemperatureModuleId(module_id),
                plate_target_temperature=command.result.targetTemperature,
            )
        elif isinstance(command.result, temperature_module.DeactivateTemperatureResult):
            self._state.substate_by_module_id[module_id] = TemperatureModuleSubState(
                module_id=TemperatureModuleId(module_id),
                plate_target_temperature=None,
            )

    def _handle_thermocycler_module_commands(
        self,
        command: Union[
            thermocycler.SetTargetBlockTemperature,
            thermocycler.DeactivateBlock,
            thermocycler.SetTargetLidTemperature,
            thermocycler.DeactivateLid,
            thermocycler.OpenLid,
            thermocycler.CloseLid,
        ],
    ) -> None:
        module_id = command.params.moduleId
        thermocycler_substate = self._state.substate_by_module_id[module_id]
        assert isinstance(
            thermocycler_substate, ThermocyclerModuleSubState
        ), f"{module_id} is not a thermocycler module."

        # Get current values to preserve target temperature not being set/deactivated
        block_temperature = thermocycler_substate.target_block_temperature
        lid_temperature = thermocycler_substate.target_lid_temperature
        is_lid_open = thermocycler_substate.is_lid_open

        if isinstance(command.result, thermocycler.SetTargetBlockTemperatureResult):
            self._state.substate_by_module_id[module_id] = ThermocyclerModuleSubState(
                module_id=ThermocyclerModuleId(module_id),
                is_lid_open=is_lid_open,
                target_block_temperature=command.result.targetBlockTemperature,
                target_lid_temperature=lid_temperature,
            )
        elif isinstance(command.result, thermocycler.DeactivateBlockResult):
            self._state.substate_by_module_id[module_id] = ThermocyclerModuleSubState(
                module_id=ThermocyclerModuleId(module_id),
                is_lid_open=is_lid_open,
                target_block_temperature=None,
                target_lid_temperature=lid_temperature,
            )
        elif isinstance(command.result, thermocycler.SetTargetLidTemperatureResult):
            self._state.substate_by_module_id[module_id] = ThermocyclerModuleSubState(
                module_id=ThermocyclerModuleId(module_id),
                is_lid_open=is_lid_open,
                target_block_temperature=block_temperature,
                target_lid_temperature=command.result.targetLidTemperature,
            )
        elif isinstance(command.result, thermocycler.DeactivateLidResult):
            self._state.substate_by_module_id[module_id] = ThermocyclerModuleSubState(
                module_id=ThermocyclerModuleId(module_id),
                is_lid_open=is_lid_open,
                target_block_temperature=block_temperature,
                target_lid_temperature=None,
            )
        elif isinstance(command.result, thermocycler.OpenLidResult):
            self._state.substate_by_module_id[module_id] = ThermocyclerModuleSubState(
                module_id=ThermocyclerModuleId(module_id),
                is_lid_open=True,
                target_block_temperature=block_temperature,
                target_lid_temperature=lid_temperature,
            )
        elif isinstance(command.result, thermocycler.CloseLidResult):
            self._state.substate_by_module_id[module_id] = ThermocyclerModuleSubState(
                module_id=ThermocyclerModuleId(module_id),
                is_lid_open=False,
                target_block_temperature=block_temperature,
                target_lid_temperature=lid_temperature,
            )

    def _handle_absorbance_reader_commands(
        self,
        command: Union[
            absorbance_reader.Initialize,
            absorbance_reader.ReadAbsorbance,
        ],
    ) -> None:
        module_id = command.params.moduleId
        absorbance_reader_substate = self._state.substate_by_module_id[module_id]
        assert isinstance(
            absorbance_reader_substate, AbsorbanceReaderSubState
        ), f"{module_id} is not an absorbance plate reader."

        # Get current values
        configured = absorbance_reader_substate.configured
        measure_mode = absorbance_reader_substate.measure_mode
        configured_wavelengths = absorbance_reader_substate.configured_wavelengths
        reference_wavelength = absorbance_reader_substate.reference_wavelength
        is_lid_on = absorbance_reader_substate.is_lid_on

        if isinstance(command.result, absorbance_reader.InitializeResult):
            self._state.substate_by_module_id[module_id] = AbsorbanceReaderSubState(
                module_id=AbsorbanceReaderId(module_id),
                configured=True,
                measured=False,
                is_lid_on=is_lid_on,
                measure_mode=AbsorbanceReaderMeasureMode(command.params.measureMode),
                configured_wavelengths=command.params.sampleWavelengths,
                reference_wavelength=command.params.referenceWavelength,
                data=None,
            )
        elif isinstance(command.result, absorbance_reader.ReadAbsorbanceResult):
            self._state.substate_by_module_id[module_id] = AbsorbanceReaderSubState(
                module_id=AbsorbanceReaderId(module_id),
                configured=configured,
                measured=True,
                is_lid_on=is_lid_on,
                measure_mode=measure_mode,
                configured_wavelengths=configured_wavelengths,
                reference_wavelength=reference_wavelength,
                data=command.result.data,
            )


class ModuleView(HasState[ModuleState]):
    """Read-only view of computed module state."""

    _state: ModuleState

    def __init__(self, state: ModuleState) -> None:
        """Initialize the view with its backing state value."""
        self._state = state

    def get(self, module_id: str) -> LoadedModule:
        """Get module data by the module's unique identifier."""
        try:
            slot_name = self._state.slot_by_module_id[module_id]
            attached_module = self._state.hardware_by_module_id[module_id]

        except KeyError as e:
            raise errors.ModuleNotLoadedError(module_id=module_id) from e

        location = (
            DeckSlotLocation(slotName=slot_name) if slot_name is not None else None
        )

        return LoadedModule.construct(
            id=module_id,
            location=location,
            model=attached_module.definition.model,
            serialNumber=attached_module.serial_number,
        )

    def get_all(self) -> List[LoadedModule]:
        """Get a list of all module entries in state."""
        return [self.get(mod_id) for mod_id in self._state.slot_by_module_id.keys()]

    def get_by_slot(
        self,
        slot_name: DeckSlotName,
    ) -> Optional[LoadedModule]:
        """Get the module located in a given slot, if any."""
        slots_by_id = reversed(list(self._state.slot_by_module_id.items()))

        for module_id, module_slot in slots_by_id:
            if module_slot == slot_name:
                return self.get(module_id)

        return None

    def _get_module_substate(
        self, module_id: str, expected_type: Type[ModuleSubStateT], expected_name: str
    ) -> ModuleSubStateT:
        """Return the specific sub-state of a given module ID.

        Args:
            module_id: The ID of the module.
            expected_type: The shape of the substate that we expect.
            expected_name: A user-friendly name of the module to put into an
                error message if the substate does not match the expected type.

        Raises:
            ModuleNotLoadedError: If module_id has not been loaded.
            WrongModuleTypeError: If module_id has been loaded,
                but it's not the expected type.
        """
        try:
            substate = self._state.substate_by_module_id[module_id]
        except KeyError as e:
            raise errors.ModuleNotLoadedError(module_id=module_id) from e

        if isinstance(substate, expected_type):
            return substate

        raise errors.WrongModuleTypeError(f"{module_id} is not a {expected_name}.")

    def get_magnetic_module_substate(self, module_id: str) -> MagneticModuleSubState:
        """Return a `MagneticModuleSubState` for the given Magnetic Module.

        Raises:
            ModuleNotLoadedError: If module_id has not been loaded.
            WrongModuleTypeError: If module_id has been loaded,
                but it's not a Magnetic Module.
        """
        return self._get_module_substate(
            module_id=module_id,
            expected_type=MagneticModuleSubState,
            expected_name="Magnetic Module",
        )

    def get_heater_shaker_module_substate(
        self, module_id: str
    ) -> HeaterShakerModuleSubState:
        """Return a `HeaterShakerModuleSubState` for the given Heater-Shaker Module.

        Raises:
           ModuleNotLoadedError: If module_id has not been loaded.
           WrongModuleTypeError: If module_id has been loaded,
               but it's not a Heater-Shaker Module.
        """
        return self._get_module_substate(
            module_id=module_id,
            expected_type=HeaterShakerModuleSubState,
            expected_name="Heater-Shaker Module",
        )

    def get_temperature_module_substate(
        self, module_id: str
    ) -> TemperatureModuleSubState:
        """Return a `TemperatureModuleSubState` for the given Temperature Module.

        Raises:
           ModuleNotLoadedError: If module_id has not been loaded.
           WrongModuleTypeError: If module_id has been loaded,
               but it's not a Temperature Module.
        """
        return self._get_module_substate(
            module_id=module_id,
            expected_type=TemperatureModuleSubState,
            expected_name="Temperature Module",
        )

    def get_thermocycler_module_substate(
        self, module_id: str
    ) -> ThermocyclerModuleSubState:
        """Return a `ThermocyclerModuleSubState` for the given Thermocycler Module.

        Raises:
           ModuleNotLoadedError: If module_id has not been loaded.
           WrongModuleTypeError: If module_id has been loaded,
               but it's not a Thermocycler Module.
        """
        return self._get_module_substate(
            module_id=module_id,
            expected_type=ThermocyclerModuleSubState,
            expected_name="Thermocycler Module",
        )

    def get_absorbance_reader_substate(
        self, module_id: str
    ) -> AbsorbanceReaderSubState:
        """Return a `AbsorbanceReaderSubState` for the given Absorbance Reader.

        Raises:
           ModuleNotLoadedError: If module_id has not been loaded.
           WrongModuleTypeError: If module_id has been loaded,
               but it's not an Absorbance Reader.
        """
        return self._get_module_substate(
            module_id=module_id,
            expected_type=AbsorbanceReaderSubState,
            expected_name="Absorbance Reader",
        )

    def get_location(self, module_id: str) -> DeckSlotLocation:
        """Get the slot location of the given module."""
        location = self.get(module_id).location
        if location is None:
            raise errors.ModuleNotOnDeckError(
                f"Module {module_id} is not loaded into a deck slot."
            )
        return location

    def get_requested_model(self, module_id: str) -> Optional[ModuleModel]:
        """Return the model by which this module was requested.

        Or, if this module was not loaded with an explicit ``loadModule`` command,
        return ``None``.

        See also `get_connected_model()`.
        """
        try:
            return self._state.requested_model_by_id[module_id]
        except KeyError as e:
            raise errors.ModuleNotLoadedError(module_id=module_id) from e

    # TODO(jbl 2023-06-20) rename this method to better reflect it's not just "connected" modules
    def get_connected_model(self, module_id: str) -> ModuleModel:
        """Return the model of the connected module.

        NOTE: This method will return the name for any module loaded, not just electronically connected ones.
            This includes the Magnetic Block.

        This can differ from `get_requested_model()` because of module compatibility.
        For example, a ``loadModule`` command might request a ``temperatureModuleV1``
        but return a ``temperatureModuleV2`` if that's what it finds actually connected
        at run time.
        """
        return self.get(module_id).model

    def get_serial_number(self, module_id: str) -> str:
        """Get the hardware serial number of the given module.

        If the underlying hardware API is simulating, this will be a dummy value
        provided by the hardware API.
        """
        module = self.get(module_id)
        if module.serialNumber is None:
            raise ModuleNotConnectedError(
                f"Expected a connected module and got a {module.model.name}"
            )
        return module.serialNumber

    def get_definition(self, module_id: str) -> ModuleDefinition:
        """Module definition by ID."""
        try:
            attached_module = self._state.hardware_by_module_id[module_id]
        except KeyError as e:
            raise errors.ModuleNotLoadedError(module_id=module_id) from e

        return attached_module.definition

    def get_dimensions(self, module_id: str) -> ModuleDimensions:
        """Get the specified module's dimensions."""
        return self.get_definition(module_id).dimensions

    def get_nominal_offset_to_child(
        self,
        module_id: str,
        # todo(mm, 2024-11-07): A method of one view taking a sibling view as an argument
        # is unusual, and may be bug-prone if the order in which the views are updated
        # matters. If we need to compute something that depends on module info and
        # addressable area info, can we do that computation in GeometryView instead of
        # here?
        addressable_areas: AddressableAreaView,
    ) -> LabwareOffsetVector:
        """Get the nominal offset from a module's location to its child labware's location.

        Includes the slot-specific transform. Does not include the child's
        Labware Position Check offset.
        """
        if (
            self.state.deck_type == DeckType.OT2_STANDARD
            or self.state.deck_type == DeckType.OT2_SHORT_TRASH
        ):
            definition = self.get_definition(module_id)
            slot = self.get_location(module_id).slotName.id

            pre_transform: NDArray[npdouble] = array(
                (
                    definition.labwareOffset.x,
                    definition.labwareOffset.y,
                    definition.labwareOffset.z,
                    1,
                )
            )
            xforms_ser = definition.slotTransforms.get(
                str(self._state.deck_type.value), {}
            ).get(
                slot,
                {
                    "labwareOffset": [
                        [1, 0, 0, 0],
                        [0, 1, 0, 0],
                        [0, 0, 1, 0],
                        [0, 0, 0, 1],
                    ]
                },
            )
            xforms_ser_offset = xforms_ser["labwareOffset"]

            # Apply the slot transform, if any
            xform: NDArray[npdouble] = array(xforms_ser_offset)
            xformed = dot(xform, pre_transform)
            return LabwareOffsetVector(
                x=xformed[0],
                y=xformed[1],
                z=xformed[2],
            )
        else:
            module = self.get(module_id)
            if isinstance(module.location, DeckSlotLocation):
                location = module.location.slotName
            elif module.model == ModuleModel.THERMOCYCLER_MODULE_V2:
                location = DeckSlotName.SLOT_B1
            else:
                raise ValueError(
                    "Module location invalid for nominal module offset calculation."
                )
            module_addressable_area = self.ensure_and_convert_module_fixture_location(
                location, self.state.deck_type, module.model
            )
            module_addressable_area_position = (
                addressable_areas.get_addressable_area_offsets_from_cutout(
                    module_addressable_area
                )
            )
            return LabwareOffsetVector(
                x=module_addressable_area_position.x,
                y=module_addressable_area_position.y,
                z=module_addressable_area_position.z,
            )

    def get_module_calibration_offset(
        self, module_id: str
    ) -> Optional[ModuleOffsetData]:
        """Get the calibration module offset."""
        module_serial = self.get(module_id).serialNumber
        if module_serial:
            return self._state.module_offset_by_serial.get(module_serial)
        return None

    def get_overall_height(self, module_id: str) -> float:
        """Get the height of the module, excluding any labware loaded atop it."""
        return self.get_dimensions(module_id).bareOverallHeight

    # TODO(mc, 2022-01-19): this method is missing unit test coverage
    def get_height_over_labware(self, module_id: str) -> float:
        """Get the height of module parts above module labware base."""
        return self.get_dimensions(module_id).overLabwareHeight

    def get_module_highest_z(
        self, module_id: str, addressable_areas: AddressableAreaView
    ) -> float:
        """Get the highest z point of the module, as placed on the robot.

        The highest Z of a module, unlike the bare overall height, depends on
        the robot it is on. We will calculate this value using the info we already have
        about the transformation of the module's placement, based on the deck it is on.

        This value is calculated as:
        highest_z = ( nominal_robot_transformed_labware_offset_z
                      + z_difference_between_default_labware_offset_point_and_overall_height
                      + module_calibration_offset_z
        )

        For OT2, the default_labware_offset point is the same as nominal_robot_transformed_labware_offset_z
        and hence the highest z will equal to the overall height of the module.

        For Flex, since those two offsets are not the same, the final highest z will be
        transformed the same amount as the labware offset point is.

        Note: For thermocycler, the lid height is not taken into account.
        """
        module_height = self.get_overall_height(module_id)
        default_lw_offset_point = self.get_definition(module_id).labwareOffset.z
        z_difference = module_height - default_lw_offset_point

        nominal_transformed_lw_offset_z = self.get_nominal_offset_to_child(
            module_id=module_id, addressable_areas=addressable_areas
        ).z
        calibration_offset = self.get_module_calibration_offset(module_id)
        return (
            nominal_transformed_lw_offset_z
            + z_difference
            + (calibration_offset.moduleOffsetVector.z if calibration_offset else 0)
        )

    # TODO(mc, 2022-01-19): this method is missing unit test coverage and
    # is also unused. Remove or add tests.
    def get_lid_height(self, module_id: str) -> float:
        """Get lid height if module is thermocycler."""
        definition = self.get_definition(module_id)

        if (
            definition.moduleType == ModuleType.THERMOCYCLER
            and hasattr(definition.dimensions, "lidHeight")
            and definition.dimensions.lidHeight is not None
        ):
            return definition.dimensions.lidHeight
        else:
            raise errors.WrongModuleTypeError(
                f"Cannot get lid height of {definition.moduleType}"
            )

    @staticmethod
    def get_magnet_home_to_base_offset(module_model: ModuleModel) -> float:
        """Return a Magnetic Module's home offset.

        This is how far a Magnetic Module's magnets have to rise above their
        home position for their tops to be level with the bottom of the labware.

        The offset is returned in true millimeters,
        even though GEN1 Magnetic Modules are sometimes controlled in units of
        half-millimeters ("short mm").
        """
        if module_model == ModuleModel.MAGNETIC_MODULE_V1:
            offset_in_half_mm = MAGNETIC_MODULE_OFFSET_TO_LABWARE_BOTTOM[
                "magneticModuleV1"
            ]
            return offset_in_half_mm / 2
        elif module_model == ModuleModel.MAGNETIC_MODULE_V2:
            return MAGNETIC_MODULE_OFFSET_TO_LABWARE_BOTTOM["magneticModuleV2"]
        else:
            raise errors.WrongModuleTypeError(
                f"Can't get magnet offset of {module_model}."
            )

    @overload
    @classmethod
    def calculate_magnet_height(
        cls,
        *,
        module_model: ModuleModel,
        height_from_home: float,
    ) -> float:
        pass

    @overload
    @classmethod
    def calculate_magnet_height(
        cls,
        *,
        module_model: ModuleModel,
        height_from_base: float,
    ) -> float:
        pass

    @overload
    @classmethod
    def calculate_magnet_height(
        cls,
        *,
        module_model: ModuleModel,
        labware_default_height: float,
        offset_from_labware_default: float,
    ) -> float:
        pass

    @classmethod
    def calculate_magnet_height(
        cls,
        *,
        module_model: ModuleModel,
        height_from_home: Optional[float] = None,
        height_from_base: Optional[float] = None,
        labware_default_height: Optional[float] = None,
        offset_from_labware_default: Optional[float] = None,
    ) -> float:
        """Normalize a Magnetic Module engage height to standard units.

        Args:
            module_model: What kind of Magnetic Module to calculate the height for.
            height_from_home: A distance above the magnets' home position,
                in millimeters.
            height_from_base: A distance above the labware base plane,
                in millimeters.
            labware_default_height: A distance above the labware base plane,
                in millimeters, from a labware definition.
            offset_from_labware_default: A distance from the
                ``labware_default_height`` argument, in hardware units.

        Negative values are allowed for all arguments, to move down instead of up.

        See the overload signatures for which combinations of parameters are allowed.

        Returns:
            The same height passed in, converted to be measured in
            millimeters above the module's labware base plane,
            suitable as input to a Magnetic Module engage Protocol Engine command.
        """
        if height_from_home is not None:
            home_to_base = cls.get_magnet_home_to_base_offset(module_model=module_model)
            return height_from_home - home_to_base

        elif height_from_base is not None:
            return height_from_base

        else:
            # Guaranteed statically by overload.
            assert labware_default_height is not None
            assert offset_from_labware_default is not None
            return labware_default_height + offset_from_labware_default

    def should_dodge_thermocycler(
        self,
        from_slot: Union[DeckSlotName, StagingSlotName],
        to_slot: Union[DeckSlotName, StagingSlotName],
    ) -> bool:
        """Decide if the requested path would cross the thermocycler, if installed.

        Returns True if we need to dodge, False otherwise.
        """
        all_mods = self.get_all()
        if any(ModuleModel.is_thermocycler_module_model(mod.model) for mod in all_mods):
            transit = (from_slot, to_slot)
            if transit in _THERMOCYCLER_SLOT_TRANSITS_TO_DODGE:
                return True
        return False

    def is_edge_move_unsafe(self, mount: MountType, target_slot: DeckSlotName) -> bool:
        """Check if the slot next to target contains a module to be avoided, depending on mount."""
        slot_int = target_slot.as_int()

        if mount is MountType.RIGHT:
            # Check left of the target
            neighbor_int = get_west_slot(slot_int)
            if neighbor_int is None:
                return False
            else:
                neighbor_slot = DeckSlotName.from_primitive(neighbor_int)
        else:
            # Check right of the target
            neighbor_int = get_east_slot(slot_int)
            if neighbor_int is None:
                return False
            else:
                neighbor_slot = DeckSlotName.from_primitive(neighbor_int)

        return neighbor_slot in self._state.slot_by_module_id.values()

    def select_hardware_module_to_load(  # noqa: C901
        self,
        model: ModuleModel,
        location: DeckSlotLocation,
        attached_modules: Sequence[HardwareModule],
        expected_serial_number: Optional[str] = None,
    ) -> HardwareModule:
        """Get the next matching hardware module for the given model and location.

        If a "matching" model is found already loaded in state at the requested
        location, that hardware module will be "reused" and selected. This behavior
        allows multiple load module commands to be issued while always preserving
        module hardware instance to deck slot mapping, which is required for
        multiples-of-a-module functionality.

        Args:
            model: The requested module model. The selected module may have a
                different model if the definition lists the model as compatible.
            location: The location the module will be assigned to.
            attached_modules: All attached modules as reported by the HardwareAPI,
                in the order in which they should be used.
            expected_serial_number: An optional variable containing the serial number
                expected of the module identified.

        Raises:
            ModuleNotAttachedError: A not-yet-assigned module matching the requested
                parameters could not be found in the attached modules list.
            ModuleAlreadyPresentError: A module of a different type is already
                assigned to the requested location.
        """
        existing_mod_in_slot = None

        for mod_id, slot in self._state.slot_by_module_id.items():
            if slot == location.slotName:
                existing_mod_in_slot = self._state.hardware_by_module_id.get(mod_id)
                break
        if existing_mod_in_slot:
            existing_def = existing_mod_in_slot.definition

            if existing_def.model == model or model in existing_def.compatibleWith:
                return existing_mod_in_slot

            else:
                raise errors.ModuleAlreadyPresentError(
                    f"A {existing_def.model.value} is already"
                    f" present in {location.slotName.value}"
                )

        for m in attached_modules:
            if m not in self._state.hardware_by_module_id.values():
                if model == m.definition.model or model in m.definition.compatibleWith:
                    if expected_serial_number is not None:
                        if m.serial_number == expected_serial_number:
                            return m
                    else:
                        return m

        raise errors.ModuleNotAttachedError(f"No available {model.value} found.")

    def get_heater_shaker_movement_restrictors(
        self,
    ) -> List[HeaterShakerMovementRestrictors]:
        """Get shaking status, latch status, and location for every heater-shaker on deck."""
        hs_substates = [
            self.get_heater_shaker_module_substate(module_id=module.id)
            for module in self.get_all()
            if module.model == ModuleModel.HEATER_SHAKER_MODULE_V1
        ]
        hs_restrictors = [
            HeaterShakerMovementRestrictors(
                plate_shaking=substate.is_plate_shaking,
                latch_status=substate.labware_latch_status,
                deck_slot=self.get_location(substate.module_id).slotName.as_int(),
            )
            for substate in hs_substates
        ]
        return hs_restrictors

    def raise_if_module_in_location(
        self,
        location: DeckSlotLocation,
    ) -> None:
        """Raise if the given location has a module in it."""
        for module in self.get_all():
            if module.location == location:
                raise errors.LocationIsOccupiedError(
                    f"Module {module.model} is already present at {location}."
                )

    def get_default_gripper_offsets(
        self, module_id: str
    ) -> Optional[LabwareMovementOffsetData]:
        """Get the deck's default gripper offsets."""
        offsets = self.get_definition(module_id).gripperOffsets
        return offsets.get("default") if offsets else None

    def get_overflowed_module_in_slot(
        self, slot_name: DeckSlotName
    ) -> Optional[LoadedModule]:
        """Get the module that's not loaded in the given slot, but still occupies the slot.

        For example, if there's a thermocycler loaded in B1,
        `get_overflowed_module_in_slot(DeckSlotName.Slot_A1)` will return the loaded
        thermocycler module.
        """
        slots_by_id = self._state.additional_slots_occupied_by_module_id

        for module_id, module_slots in slots_by_id.items():
            if module_slots and slot_name in module_slots:
                return self.get(module_id)

        return None

    def is_flex_deck_with_thermocycler(self) -> bool:
        """Return if this is a Flex deck with a thermocycler loaded in B1-A1 slots."""
        maybe_module = self.get_by_slot(
            DeckSlotName.SLOT_A1
        ) or self.get_overflowed_module_in_slot(DeckSlotName.SLOT_A1)
        if (
            self._state.deck_type == DeckType.OT3_STANDARD
            and maybe_module
            and maybe_module.model == ModuleModel.THERMOCYCLER_MODULE_V2
        ):
            return True
        else:
            return False

    def convert_absorbance_reader_data_points(
        self, data: List[float]
    ) -> Dict[str, float]:
        """Return the data from the Absorbance Reader module in a map of wells for each read value."""
        if len(data) == 96:
            # We have to reverse the reader values because the Opentrons Absorbance Reader is rotated 180 degrees on the deck
            data.reverse()
            well_map: Dict[str, float] = {}
            for i, value in enumerate(data):
                row = chr(ord("A") + i // 12)  # Convert index to row (A-H)
                col = (i % 12) + 1  # Convert index to column (1-12)
                well_key = f"{row}{col}"
                truncated_value = float(
                    "{:.5}".format(str(value))
                )  # Truncate the returned value to the third decimal place
                well_map[well_key] = truncated_value
            return well_map
        else:
            raise ValueError(
                "Only readings of 96 Well labware are supported for conversion to map of values by well."
            )

    def ensure_and_convert_module_fixture_location(
        self,
        deck_slot: DeckSlotName,
        deck_type: DeckType,
        model: ModuleModel,
    ) -> str:
        """Ensure module fixture load location is valid.

        Also, convert the deck slot to a valid module fixture addressable area.
        """
        if deck_type == DeckType.OT2_STANDARD or deck_type == DeckType.OT2_SHORT_TRASH:
            raise ValueError(
                f"Invalid Deck Type: {deck_type.name} - Does not support modules as fixtures."
            )

        assert deck_slot in DeckSlotName.ot3_slots()
        if model == ModuleModel.MAGNETIC_BLOCK_V1:
            return f"magneticBlockV1{deck_slot.value}"

        elif model == ModuleModel.HEATER_SHAKER_MODULE_V1:
            # only allowed in column 1 & 3
            assert deck_slot.value[-1] in ("1", "3")
            return f"heaterShakerV1{deck_slot.value}"

        elif model == ModuleModel.TEMPERATURE_MODULE_V2:
            # only allowed in column 1 & 3
            assert deck_slot.value[-1] in ("1", "3")
            return f"temperatureModuleV2{deck_slot.value}"

        elif model == ModuleModel.THERMOCYCLER_MODULE_V2:
            return "thermocyclerModuleV2"

        elif model == ModuleModel.ABSORBANCE_READER_V1:
            # only allowed in column 3
            assert deck_slot.value[-1] == "3"
            return f"absorbanceReaderV1{deck_slot.value}"

        raise ValueError(
            f"Unknown module {model.name} has no addressable areas to provide."
        )

    def absorbance_reader_dock_location(
        self, module_id: str
    ) -> AddressableAreaLocation:
        """Get the addressable area for the absorbance reader dock."""
        reader_slot = self.get_location(module_id)
        lid_doc_slot = get_adjacent_staging_slot(reader_slot.slotName)
        assert lid_doc_slot is not None
        lid_dock_area = AddressableAreaLocation(
            addressableAreaName="absorbanceReaderV1LidDock" + lid_doc_slot.value
        )
        return lid_dock_area



"""Models and implementation for the ``moveLabware`` command."""

from __future__ import annotations
from opentrons_shared_data.errors.exceptions import (
    FailedGripperPickupError,
    LabwareDroppedError,
    StallOrCollisionDetectedError,
)
from pydantic import BaseModel, Field
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from opentrons.protocol_engine.resources.model_utils import ModelUtils
from opentrons.types import Point
from ..types import (
    ModuleModel,
    CurrentWell,
    LabwareLocation,
    DeckSlotLocation,
    ModuleLocation,
    OnLabwareLocation,
    AddressableAreaLocation,
    LabwareMovementStrategy,
    LabwareOffsetVector,
    LabwareMovementOffsetData,
)
from ..errors import (
    LabwareMovementNotAllowedError,
    NotSupportedOnRobotType,
    LabwareOffsetDoesNotExistError,
)
from ..resources import labware_validation, fixture_validation
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)
from ..errors.error_occurrence import ErrorOccurrence
from ..state.update_types import StateUpdate
from opentrons_shared_data.gripper.constants import GRIPPER_PADDLE_WIDTH

if TYPE_CHECKING:
    from ..execution import EquipmentHandler, RunControlHandler, LabwareMovementHandler
    from ..state.state import StateView


MoveLabwareCommandType = Literal["moveLabware"]


# Extra buffer on top of minimum distance to move to the right
_TRASH_CHUTE_DROP_BUFFER_MM = 8


class MoveLabwareParams(BaseModel):
    """Input parameters for a ``moveLabware`` command."""

    labwareId: str = Field(..., description="The ID of the labware to move.")
    newLocation: LabwareLocation = Field(..., description="Where to move the labware.")
    strategy: LabwareMovementStrategy = Field(
        ...,
        description="Whether to use the gripper to perform the labware movement"
        " or to perform a manual movement with an option to pause.",
    )
    pickUpOffset: Optional[LabwareOffsetVector] = Field(
        None,
        description="Offset to use when picking up labware. "
        "Experimental param, subject to change",
    )
    dropOffset: Optional[LabwareOffsetVector] = Field(
        None,
        description="Offset to use when dropping off labware. "
        "Experimental param, subject to change",
    )


class MoveLabwareResult(BaseModel):
    """The output of a successful ``moveLabware`` command."""

    offsetId: Optional[str] = Field(
        # Default `None` instead of `...` so this field shows up as non-required in
        # OpenAPI. The server is allowed to omit it or make it null.
        None,
        description=(
            "An ID referencing the labware offset that will apply to this labware"
            " now that it's in the new location."
            " This offset will be in effect until the labware is moved"
            " with another `moveLabware` command."
            " Null or undefined means no offset applies,"
            " so the default of (0, 0, 0) will be used."
        ),
    )


class GripperMovementError(ErrorOccurrence):
    """Returned when something physically goes wrong when the gripper moves labware.

    When this error happens, the engine will leave the labware in its original place.
    """

    isDefined: bool = True

    errorType: Literal["gripperMovement"] = "gripperMovement"


_ExecuteReturn = SuccessData[MoveLabwareResult] | DefinedErrorData[GripperMovementError]


class MoveLabwareImplementation(AbstractCommandImpl[MoveLabwareParams, _ExecuteReturn]):
    """The execution implementation for ``moveLabware`` commands."""

    def __init__(
        self,
        model_utils: ModelUtils,
        state_view: StateView,
        equipment: EquipmentHandler,
        labware_movement: LabwareMovementHandler,
        run_control: RunControlHandler,
        **kwargs: object,
    ) -> None:
        self._model_utils = model_utils
        self._state_view = state_view
        self._equipment = equipment
        self._labware_movement = labware_movement
        self._run_control = run_control

    async def execute(self, params: MoveLabwareParams) -> _ExecuteReturn:  # noqa: C901
        """Move a loaded labware to a new location."""
        state_update = StateUpdate()

        # Allow propagation of LabwareNotLoadedError.
        current_labware = self._state_view.labware.get(labware_id=params.labwareId)
        current_labware_definition = self._state_view.labware.get_definition(
            labware_id=params.labwareId
        )
        definition_uri = current_labware.definitionUri
        post_drop_slide_offset: Optional[Point] = None
        trash_lid_drop_offset: Optional[LabwareOffsetVector] = None

        if self._state_view.labware.is_fixed_trash(params.labwareId):
            raise LabwareMovementNotAllowedError(
                f"Cannot move fixed trash labware '{current_labware_definition.parameters.loadName}'."
            )

        if isinstance(params.newLocation, AddressableAreaLocation):
            area_name = params.newLocation.addressableAreaName
            if (
                not fixture_validation.is_gripper_waste_chute(area_name)
                and not fixture_validation.is_deck_slot(area_name)
                and not fixture_validation.is_trash(area_name)
            ):
                raise LabwareMovementNotAllowedError(
                    f"Cannot move {current_labware.loadName} to addressable area {area_name}"
                )
            self._state_view.addressable_areas.raise_if_area_not_in_deck_configuration(
                area_name
            )

            if fixture_validation.is_gripper_waste_chute(area_name):
                # When dropping off labware in the waste chute, some bigger pieces
                # of labware (namely tipracks) can get stuck between a gripper
                # paddle and the bottom of the waste chute, even after the gripper
                # has homed all the way to the top of its travel. We add a "post-drop
                # slide" to dropoffs in the waste chute in order to guarantee that the
                # labware can drop fully through the chute before the gripper jaws close.
                post_drop_slide_offset = Point(
                    x=(current_labware_definition.dimensions.xDimension / 2.0)
                    + (GRIPPER_PADDLE_WIDTH / 2.0)
                    + _TRASH_CHUTE_DROP_BUFFER_MM,
                    y=0,
                    z=0,
                )
            elif fixture_validation.is_trash(area_name):
                # When dropping labware in the trash bins we want to ensure they are lids
                # and enforce a y-axis drop offset to ensure they fall within the trash bin
                if labware_validation.validate_definition_is_lid(
                    self._state_view.labware.get_definition(params.labwareId)
                ):
                    lid_disposable_offfets = (
                        current_labware_definition.gripperOffsets.get(
                            "lidDisposalOffsets"
                        )
                    )
                    if lid_disposable_offfets is not None:
                        trash_lid_drop_offset = LabwareOffsetVector(
                            x=lid_disposable_offfets.dropOffset.x,
                            y=lid_disposable_offfets.dropOffset.y,
                            z=lid_disposable_offfets.dropOffset.z,
                        )
                    else:
                        raise LabwareOffsetDoesNotExistError(
                            f"Labware Definition {current_labware.loadName} does not contain required field 'lidDisposalOffsets' of 'gripperOffsets'."
                        )
                else:
                    raise LabwareMovementNotAllowedError(
                        "Can only move labware with allowed role 'Lid' to a Trash Bin."
                    )

        elif isinstance(params.newLocation, DeckSlotLocation):
            self._state_view.addressable_areas.raise_if_area_not_in_deck_configuration(
                params.newLocation.slotName.id
            )

        available_new_location = self._state_view.geometry.ensure_location_not_occupied(
            location=params.newLocation
        )

        # Check that labware and destination do not have labware on top
        self._state_view.labware.raise_if_labware_has_labware_on_top(
            labware_id=params.labwareId
        )
        if isinstance(available_new_location, OnLabwareLocation):
            self._state_view.labware.raise_if_labware_has_labware_on_top(
                available_new_location.labwareId
            )
            # Ensure that labware can be placed on requested labware
            self._state_view.labware.raise_if_labware_cannot_be_stacked(
                top_labware_definition=current_labware_definition,
                bottom_labware_id=available_new_location.labwareId,
            )
            if params.labwareId == available_new_location.labwareId:
                raise LabwareMovementNotAllowedError(
                    "Cannot move a labware onto itself."
                )
        # Validate labware for the absorbance reader
        elif isinstance(available_new_location, ModuleLocation):
            module = self._state_view.modules.get(available_new_location.moduleId)
            if module is not None and module.model == ModuleModel.ABSORBANCE_READER_V1:
                self._state_view.labware.raise_if_labware_incompatible_with_plate_reader(
                    current_labware_definition
                )

        # Allow propagation of ModuleNotLoadedError.
        new_offset_id = self._equipment.find_applicable_labware_offset_id(
            labware_definition_uri=definition_uri,
            labware_location=available_new_location,
        )
        await self._labware_movement.ensure_movement_not_obstructed_by_module(
            labware_id=params.labwareId, new_location=available_new_location
        )

        if params.strategy == LabwareMovementStrategy.USING_GRIPPER:
            if self._state_view.config.robot_type == "OT-2 Standard":
                raise NotSupportedOnRobotType(
                    message="Labware movement using a gripper is not supported on the OT-2",
                    details={"strategy": params.strategy},
                )
            if not labware_validation.validate_gripper_compatible(
                current_labware_definition
            ):
                raise LabwareMovementNotAllowedError(
                    f"Cannot move labware '{current_labware_definition.parameters.loadName}' with gripper."
                    f" If trying to move a labware on an adapter, load the adapter separately to allow"
                    f" gripper movement."
                )
            if labware_validation.validate_definition_is_adapter(
                current_labware_definition
            ):
                raise LabwareMovementNotAllowedError(
                    f"Cannot move adapter '{current_labware_definition.parameters.loadName}' with gripper."
                )

            validated_current_loc = (
                self._state_view.geometry.ensure_valid_gripper_location(
                    current_labware.location
                )
            )
            validated_new_loc = self._state_view.geometry.ensure_valid_gripper_location(
                available_new_location,
            )
            user_offset_data = LabwareMovementOffsetData(
                pickUpOffset=params.pickUpOffset or LabwareOffsetVector(x=0, y=0, z=0),
                dropOffset=params.dropOffset or LabwareOffsetVector(x=0, y=0, z=0),
            )

            if trash_lid_drop_offset:
                user_offset_data.dropOffset += trash_lid_drop_offset

            try:
                # Skips gripper moves when using virtual gripper
                await self._labware_movement.move_labware_with_gripper(
                    labware_id=params.labwareId,
                    current_location=validated_current_loc,
                    new_location=validated_new_loc,
                    user_offset_data=user_offset_data,
                    post_drop_slide_offset=post_drop_slide_offset,
                )
            except (
                FailedGripperPickupError,
                LabwareDroppedError,
                StallOrCollisionDetectedError,
                # todo(mm, 2024-09-26): Catch LabwareNotPickedUpError when that exists and
                # move_labware_with_gripper() raises it.
            ) as exception:
                gripper_movement_error: GripperMovementError | None = (
                    GripperMovementError(
                        id=self._model_utils.generate_id(),
                        createdAt=self._model_utils.get_timestamp(),
                        errorCode=exception.code.value.code,
                        detail=exception.code.value.detail,
                        wrappedErrors=[
                            ErrorOccurrence.from_failed(
                                id=self._model_utils.generate_id(),
                                createdAt=self._model_utils.get_timestamp(),
                                error=exception,
                            )
                        ],
                    )
                )
            else:
                gripper_movement_error = None

            # All mounts will have been retracted as part of the gripper move.
            state_update.clear_all_pipette_locations()

            if gripper_movement_error:
                return DefinedErrorData(
                    public=gripper_movement_error,
                    state_update=state_update,
                )

        elif params.strategy == LabwareMovementStrategy.MANUAL_MOVE_WITH_PAUSE:
            # Pause to allow for manual labware movement
            await self._run_control.wait_for_resume()

        # We may have just moved the labware that contains the current well out from
        # under the pipette. Clear the current location to reflect the fact that the
        # pipette is no longer over any labware. This is necessary for safe path
        # planning in case the next movement goes to the same labware (now in a new
        # place).
        pipette_location = self._state_view.pipettes.get_current_location()
        if (
            isinstance(pipette_location, CurrentWell)
            and pipette_location.labware_id == params.labwareId
        ):
            state_update.clear_all_pipette_locations()

        state_update.set_labware_location(
            labware_id=params.labwareId,
            new_location=available_new_location,
            new_offset_id=new_offset_id,
        )

        return SuccessData(
            public=MoveLabwareResult(offsetId=new_offset_id),
            state_update=state_update,
        )


class MoveLabware(
    BaseCommand[MoveLabwareParams, MoveLabwareResult, GripperMovementError]
):
    """A ``moveLabware`` command."""

    commandType: MoveLabwareCommandType = "moveLabware"
    params: MoveLabwareParams
    result: Optional[MoveLabwareResult]

    _ImplementationCls: Type[MoveLabwareImplementation] = MoveLabwareImplementation


class MoveLabwareCreate(BaseCommandCreate[MoveLabwareParams]):
    """A request to create a ``moveLabware`` command."""

    commandType: MoveLabwareCommandType = "moveLabware"
    params: MoveLabwareParams

    _CommandCls: Type[MoveLabware] = MoveLabware



""" Type definitions for modules in this tree """
from dataclasses import dataclass
from typing import Any, Dict, List, NamedTuple, Optional
from enum import Enum


class MoveSplit(NamedTuple):
    split_distance: float
    split_current: float
    split_speed: float
    after_time: float
    fullstep: bool


MoveSplits = Dict[str, MoveSplit]
#: Dict mapping axes to their split parameters


@dataclass
class Temperature:
    """Tempdeck, thermocycler plate, and heater-shaker temperatures"""

    current: float
    target: Optional[float]


@dataclass
class RPM:
    """Heater-shaker plate RPM"""

    current: int
    target: Optional[int]


class HeaterShakerLabwareLatchStatus(str, Enum):
    """Heater-shaker labware latch status"""

    OPENING = "opening"
    IDLE_OPEN = "idle_open"
    CLOSING = "closing"
    IDLE_CLOSED = "idle_closed"
    IDLE_UNKNOWN = "idle_unknown"
    UNKNOWN = "unknown"


@dataclass
class PlateTemperature(Temperature):
    """Thermocycler lid temperature"""

    hold: Optional[float]


class ThermocyclerLidStatus(str, Enum):
    """Thermocycler lid status."""

    UNKNOWN = "unknown"
    CLOSED = "closed"
    IN_BETWEEN = "in_between"
    OPEN = "open"
    MAX = "max"


class AbsorbanceReaderLidStatus(str, Enum):
    """Absorbance reader lid status."""

    UNKNOWN = "unknown"
    ON = "on"
    OFF = "off"


class AbsorbanceReaderPlatePresence(str, Enum):
    """Absorbance reader plate presence."""

    UNKNOWN = "unknown"
    PRESENT = "present"
    ABSENT = "absent"


class AbsorbanceReaderDeviceState(str, Enum):
    """Absorbance reader device state."""

    UNKNOWN = "unknown"
    OK = "ok"
    BROKEN_FW = "broken_fw"
    ERROR = "error"


class ABSMeasurementMode(Enum):
    """The current mode configured for reading the Absorbance Reader."""

    SINGLE = "single"
    MULTI = "multi"


@dataclass
class ABSMeasurementConfig:
    measure_mode: ABSMeasurementMode
    sample_wavelengths: List[int]
    reference_wavelength: Optional[int]

    @property
    def data(self) -> Dict[str, Any]:
        return {
            "measureMode": self.measure_mode.value,
            "sampleWavelengths": self.sample_wavelengths,
            "referenceWavelength": self.reference_wavelength,
        }
